/*********************************************************************************
 
	The Atari AVG Vector Processor Emulator in Java

	by 

	Norbert Kehrer

	August 1998

**********************************************************************************

	Information sources:
		- Asteroids emulator of the MAME project - Thanks for the sources
		- Asteroids emulator by Chris Pile       - Thanks for the documentation

	
	The Atari AVG is a vector engine used in Atari games like Gravitar

**********************************************************************************

Atari Analog Vector Generator Instruction Set

Compiled from Atari schematics and specifications
Eric Smith  7/2/92
---------------------------------------------

NOTE: The vector generator is little-endian.  The instructions are 16 bit
      words, which need to be stored with the least significant byte in the
      lower (even) address.  They are shown here with the MSB on the left.

The stack allows four levels of subroutine calls in the TTL version, but only
three levels in the gate array version.

inst  bit pattern          description
----  -------------------  -------------------
VCTR  000- yyyy yyyy yyyy  normal vector
      zzz- xxxx xxxx xxxx
HALT  001- ---- ---- ----  halt - does CNTR also on newer hardware
SVEC  010y yyyy zzzx xxxx  short vector - don't use zero length
STAT  0110 ---- zzzz cccc  status
SCAL  0111 -bbb llll llll  scaling
CNTR  100- ---- dddd dddd  center
JSRL  101a aaaa aaaa aaaa  jump to subroutine
RTSL  110- ---- ---- ----  return
JMPL  111a aaaa aaaa aaaa  jump

-     unused bits
x, y  relative x and y coordinates in two's complement (5 or 13 bit,
      5 bit quantities are scaled by 2, so x=1 is really a length 2 vector.
z     intensity, 0 = blank, 1 means use z from STAT instruction,  2-7 are
      doubled for actual range of 4-14
c     color
b     binary scaling, multiplies all lengths by 2**(1-b), 0 is double size,
      1 is normal, 2 is half, 3 is 1/4, etc.
l     linear scaling, multiplies all lengths by 1-l/256, don't exceed $80
d     delay time, use $40
a     address (word address relative to base of vector memory)

Notes:

Quantum:
        the VCTR instruction has a four bit Z field, that is not
        doubled.  The value 2 means use Z from STAT instruction.

        the SVEC instruction can't be used

Major Havoc:
        SCAL bit 11 is used for setting a Y axis window.

        STAT bit 11 is used to enable "sparkle" color.
        STAT bit 10 inverts the X axis of vectors.
        STAT bits 9 and 8 are the Vector ROM bank select.

Star Wars:
        STAT bits 10, 9, and 8 are used directly for R, G, and B.


*/


import java.awt.*;
import java.awt.image.*;
import java.awt.image.MemoryImageSource;
import java.awt.image.ColorModel;
import java.util.*;
import java.io.*;
import java.net.*;
import java.applet.*;
import java.net.*;


// ***
// *** ATARI AVG vector engine
// ***

public class AtariAVG extends Object {

	/* AVG memory and video hardware information */
	public int[] 	avgMem;
	public int	avgVectorRAM;
	public Graphics	avgBufferGraphics;

	static int xSize; 
	static int ySize;
	static int xCenter;
	static int yCenter;
	
	static Color[] avgColorTable = new Color[256];


	/* AVG constants */
	static int MAXSTACK = 8;
	static int CORR_SHIFT = 16;
	static int INITIAL_SCALE = 1<<16;

	/* AVG commands */
	final static int VCTR = 0x00;
	final static int HALT = 0x01;
	final static int SVEC = 0x02;
	final static int STAT = 0x03;
	final static int CNTR = 0x04;
	final static int JSRL = 0x05;
	final static int RTSL = 0x06;
	final static int JMPL = 0x07;
	final static int SCAL = 0x08;


	/* Chip Registers */
	static int pc;
	static int sp;
	static int[] stack = new int [MAXSTACK];

	static long scale;
	static int statz;
	static long currentx, currenty;
	static int color;



	// *** AVG generation 

	public AtariAVG (int [] memory, int vectorRAM, Graphics bufferGr, int xs, int ys) {
		avgMem = 		memory;
		avgVectorRAM = 		vectorRAM;
		avgBufferGraphics = 	bufferGr;
		xSize =			xs;
		ySize =			ys;
		xCenter =		(int) ((xSize << CORR_SHIFT) / 2);
		yCenter =		(int) ((ySize << CORR_SHIFT) / 2);
		
		// Generate colors
		for (int c = 0; c < 8; c++) { // Color
			for (int i = 0; i < 16; i++) { // Intensity
				avgColorTable[(i<<4)+c] = new Color(
							((c & 0x04) >> 2) * (i<<4), // Red
							((c & 0x02) >> 1) * (i<<4), // Green
							 (c & 0x01)       * (i<<4)  // Blue
				);
				avgColorTable[(i<<4)+c+8] = avgColorTable[i];
			}
		}


	}
	
	
	// *** Utility functions
	
	final int readWordVectorRAM (int offset ) {
		int avgAddr = avgVectorRAM + (offset<<1);
		return ( avgMem[avgAddr] | ( avgMem[avgAddr+1] << 8) );
	}


	final int twosCompVal12(int num) {
		num &= 0xfff;
		return ( ((num&0x800) != 0) 	? (-(0x1000 - num))
						: num
		);
	}

	final int twosCompVal5(int num) {
		num &= 0x1f;
		return ( ((num&0x10) != 0) 	? (-(0x20 - num))
						: num
		);
	}



	// *** Vector drawing

	final void drawVector(long x, long y, long dx, long dy, int color, int intensity) {
		int x1,y1,x2,y2;
		if (intensity > 0) {
			x1 = (int)(x>>CORR_SHIFT);
			y1 = (int)(ySize-(y>>CORR_SHIFT));
			x2 = (int)((x+dx)>>CORR_SHIFT);
			y2 = (int)(ySize-((y+dy)>>CORR_SHIFT));
			avgBufferGraphics.setColor(avgColorTable[(intensity<<4)|color]);
			avgBufferGraphics.drawLine(x1, y1, x2, y2);
		}
	}


	// *** Reset the AVG 
	
	public final void avgReset() {

		pc = 0;
		sp = 0;
		scale = INITIAL_SCALE;
		statz = 0;
		color = 0;
		
		currentx = (long) xCenter;
		currenty = (long) yCenter;
	}


	// *** Execute the AVG commands
	
	public final void avgRun() {

		boolean done = false;

		int firstwd  = 0;
		int secondwd = 0;
		int opcode;

		long x, y;
		int z = 0;
		int l, b, d, a;

		long deltax, deltay;


		// *** Clear Screen

		avgBufferGraphics.setColor (Color.black);
		avgBufferGraphics.fillRect(0,0,xSize,ySize);


		// *** Initialize state machine

		pc = 0;
		sp = 0;
		scale = INITIAL_SCALE;
		statz = 0;
		color = 0;
		
		currentx = xCenter;
		currenty = yCenter;


		// *** Execute the AVG commands

		while (!done) {

			firstwd = readWordVectorRAM(pc);
			opcode = firstwd >> 13;
			
			pc++;

			if ( opcode == VCTR ) {
				secondwd = readWordVectorRAM (pc);
				pc++;
			}

			if ((opcode == STAT) && ((firstwd & 0x1000) != 0)) opcode = SCAL;


			switch (opcode) {

			case VCTR:
				x = twosCompVal12(secondwd);
				y = twosCompVal12(firstwd);

				z = (secondwd >> 12) & 0x0e;
				if (z == 2) z = statz;

				deltax = x * scale;
				deltay = y * scale;

				drawVector(currentx, currenty, deltax, deltay, color, z); 

	  			currentx += deltax;
				currenty += deltay;
								
				break;

			case SVEC:
				x = twosCompVal5(firstwd) * 2;
				y = twosCompVal5(firstwd >> 8) * 2 ;
				z = ((firstwd >> 4) & 0x0e);
				if (z == 2) z = statz;

				deltax = x * scale;
				deltay = y * scale;

				drawVector(currentx, currenty, deltax, deltay, color, z); 
				currentx += deltax;
				currenty += deltay;
				break;



			case STAT:
				color = (firstwd) & 0x000f;
				statz = (firstwd >> 4) & 0x000f;
				break;


			case SCAL:
				b = ((firstwd >> 8) & 0x07)+8;
				l = 0x100 - (firstwd & 0xff);
				scale = (l << CORR_SHIFT) >> b;
				break;


			case CNTR:
				d = firstwd & 0xff;
				currentx = xCenter ;
				currenty = yCenter ;
				break;

			case RTSL:
				if (sp == 0) {
					System.out.println("\n*** Vector generator stack overflow! ***\n");
					done = true;
					sp = MAXSTACK - 1;
				}
				else
					sp--;
				pc = stack [sp];

				break;

			case HALT:
				done = true;
				break;

			case JMPL:
				a = firstwd & 0x1fff;
				if (a == 0x0000) 	done = true; //treat as HALT
				else			pc = a;
				break;

			case JSRL:
				a = firstwd & 0x1fff;
				if (a == 0x0000)
					done = true; //treat as HALT
				else {
					stack [sp] = pc;
					if (sp == (MAXSTACK - 1)) {
						System.out.println("\n*** Vector generator stack overflow! ***\n");
						done = true;
						sp = 0;
					}
					else
						sp++;
					
					pc = a;
				}

				break;

			default:
				System.out.println("Unknown AVG opcode found:"+opcode);
				done = true;
			}
		}


	}
}








