/*********************************************************************************
 
	The Atari DVG Vector Processor Emulator in Java

	by 

	Norbert Kehrer

	August 1998

**********************************************************************************

	Information sources:
		- Asteroids emulator of the MAME project - Thanks for the sources
		- Asteroids emulator by Chris Pile       - Thanks for the documentation

	
	The Atari DVG is a vector engine used in Atari games like Asteroids

*/


import java.awt.*;
import java.awt.image.*;
import java.awt.image.MemoryImageSource;
import java.awt.image.ColorModel;
import java.util.*;
import java.io.*;
import java.net.*;
import java.applet.*;
import java.net.*;


// ***
// *** ATARI DVG vector engine
// ***

public class AtariDVG extends Object {

	/* DVG memory and video hardware information */
	public int[] 	dvgMem;
	public int	dvgVectorRAM;
	public Graphics	dvgBufferGraphics;

	static int xSize;
	static int ySize;
	
	static Color[] dvgColorTable = new Color[16];


	/* DVG stack size */
	static int MAXSTACK = 8;

	/* DVG commands */
	static int DVCTR = 0x01;
	static int DLABS = 0x0a;
	static int DHALT = 0x0b;
	static int DJSRL = 0x0c;
	static int DRTSL = 0x0d;
	static int DJMPL = 0x0e;
	static int DSVEC = 0x0f;


	/* Chip Registers */
	static int pc;
	static int sp;
	static int[] stack = new int [MAXSTACK];

	static int scale;
	static int currentx, currenty;



	// *** DVG generation 

	public AtariDVG (int [] memory, int vectorRAM, Graphics bufferGr, int xs, int ys) {
		dvgMem = 		memory;
		dvgVectorRAM = 		vectorRAM;
		dvgBufferGraphics = 	bufferGr;
		xSize =			xs;
		ySize =			ys;
		
		// Generate colors
		for (int i = 0; i < 16; i++) dvgColorTable[i] = new Color(i<<4, i<<4, i<<4);

	}
	
	
	// *** Utility functions
	
	final int readWordVectorRAM (int offset ) {
		int dvgAddr = dvgVectorRAM + (offset<<1);
		return ( dvgMem[dvgAddr] | ( dvgMem[dvgAddr+1] << 8) );
	}

	final int twosCompVal( int num, int bits) {
		num &= 0x0fff;
		return ( ((num&0x0800) != 0) 	? (num | ~(0x0fff))
						: (num & 0xfff)
		);
	}



	// *** Vector drawing

	final void drawVector(int x, int y, int dx, int dy, int intensity) {
		int x1,y1,x2,y2,corr=3;
		if (intensity > 0) {
			x1 = (x>>corr);
			y1 = xSize-(y>>corr);
			x2 = ((x+dx)>>corr);
			y2 = ySize-((y+dy)>>corr);
			dvgBufferGraphics.setColor(dvgColorTable[intensity>>4]);
			dvgBufferGraphics.drawLine(x1, y1, x2, y2);
		}
	}


	// *** Execute the DVG commands
	
	public final void dvgRun() {

		boolean done = false;

		int firstwd  = 0;
		int secondwd = 0;
		int opcode;

		int x, y;
		int z, temp;
		int a;

		int deltax, deltay;


		// *** Clear Screen

		dvgBufferGraphics.setColor (Color.black);
		dvgBufferGraphics.fillRect(0,0,xSize,ySize);


		// *** Initialize state machine

		pc = 0;
		sp = 0;
		scale = 0;
		
		currentx = 0;
		currenty = 0;


		// *** Execute the DVG commands

		while (!done) {

			firstwd = readWordVectorRAM(pc);
			opcode = firstwd >> 12;
			
			pc++;

			if ((opcode >= DVCTR ) && (opcode <= DLABS)) {
				secondwd = readWordVectorRAM (pc);
				pc++;
			}


			switch (opcode) {

			case 0: // DHALT
				done = true;
				break;
			case 1:
			case 2:
			case 3:
			case 4:
			case 5:
			case 6:
			case 7:
			case 8:
			case 9: // DLVEC
	  			y = firstwd & 0x03ff;
				if ((firstwd & 0x0400) != 0) y = -y;
				x = secondwd & 0x03ff;
				if ((secondwd & 0x0400) != 0)	x = -x;
				z = secondwd >> 12;

	  			temp = ((scale + opcode) & 0x0f);

	  			if (temp > 9) temp = -1;
	  			deltax = x >> (7-temp);
				deltay = y >> (7-temp);

				if (z != 0) z = (z << 4) | 0x0f;

				drawVector(currentx, currenty, deltax, deltay, z); 

	  			currentx += deltax;
				currenty += deltay;
								
				break;

			case 0x0a: // DLABS
				x = twosCompVal (secondwd, 12);
				y = twosCompVal (firstwd, 12);
	  			scale = (secondwd >> 12);
				currentx = x<<2;
				currenty = y<<2;

				break;

			case 0x0b: // DHALT
				done = true;
				break;

			case 0x0c: // DJSRL
				a = firstwd & 0x0fff;
			
				stack [sp] = pc;
				if (sp == (MAXSTACK - 1)) {
					System.out.println("\n*** Vector generator stack overflow! ***\n");
					done = true;
					sp = 0;
				}
				else
					sp++;
					
				pc = a;
				break;

			case 0x0d: // DRTSL
				if (sp == 0) {
					System.out.println("\n*** Vector generator stack overflow! ***\n");
					done = true;
					sp = MAXSTACK - 1;
				}
				else
					sp--;
					
				pc = stack [sp];
				break;

			case 0x0e: // DJMPL
				a = firstwd & 0x0fff;
				pc = a;
				break;

			case 0x0f: // DSVEC
				y = firstwd & 0x0300;
				if ((firstwd & 0x0400) != 0) y = -y;
				x = (firstwd & 0x03) << 8;
				if ((firstwd & 0x04) != 0) x = -x;
				z = (firstwd >> 4) & 0x0f;
				temp = 2 + ((firstwd >> 2) & 0x02) + ((firstwd >>11) & 0x01);
	  			temp = ((scale + temp) & 0x0f);

				if (temp > 9) temp = -1;

				deltax = x >> (7-temp);
				deltay = y >> (7-temp);
	  			
				if (z != 0) z = (z << 4) | 0x0f;

				drawVector(currentx, currenty, deltax, deltay, z); 

				currentx += deltax;
				currenty += deltay;
				
				break;

			default:
				System.out.println("Unknown DVG opcode found!");
				done = true;
			}
		}


	}
}








