/*
 * @(#)J6502 - A 6502 emulator in Java */

/**
 * <p>The J6502 class emulates the 6502 microprocessor.</p>
 *
 * @version 1.0 May 1998
 * @author Norbert Kehrer</A>
 *
 * @see 6502
 * @see Emulation
 *
 * 
 * Helpful were 
 *    - M6502  by Marat Fayzullin & Alex Krasivsky
 *    - Jasper by Adam Davidson & Andrew Pollard
 **/


public class J6502 extends Object {

	public J6502( double clockFrequencyInMHz ) {
		IPeriod = (int) ((clockFrequencyInMHz * 1e6) / 50);
	}

	protected int IPeriod = 0;

	private int     ICount     = 0;
	public  int     IRequest   = 0;
	private boolean AfterCLI   = false;
	private int     IBackup    = 0;
	private int	LastRdAddr = 0;

	public  boolean Debugging  = false;


       /** Loop6502() returns:        */ 
	public final int INT_NONE = 0;	/* No interrupt required      */ 
	public final int INT_IRQ  = 1;  /* Standard IRQ interrupt     */ 
	public final int INT_NMI  = 2;  /* Non-maskable interrupt     */ 
	public final int INT_QUIT = 3;  /* Exit the emulation         */ 

 
       /** 6502 status flags:         */ 
	private final int C_FLAG = 0x01;	/* 1: Carry occured           */ 
	private final int Z_FLAG = 0x02;        /* 1: Result is zero          */ 
	private final int I_FLAG = 0x04;        /* 1: Interrupts disabled     */ 
	private final int D_FLAG = 0x08;	/* 1: Decimal mode            */ 
	private final int B_FLAG = 0x10;        /* Break [0 on stk after int] */ 
	private final int R_FLAG = 0x20;        /* Always 1                   */ 
	private final int V_FLAG = 0x40;        /* 1: Overflow occured        */ 
	private final int N_FLAG = 0x80;        /* 1: Result is negative      */ 


	/** Registers */
	protected int A = 0, P = 0, X = 0, Y = 0, S = 0;

	/** Program counter */
	public int PC = 0;




	/** Tables used by 6502 emulation for timing and to **/ 
	/** compute NEGATIVE and ZERO flags.                **/ 
 
	private int Cycles[] = 
	{ 
	 7,6,2,8,3,3,5,5,3,2,2,2,4,4,6,6, 
	 2,5,2,8,4,4,6,6,2,4,2,7,5,5,7,7, 
	 6,6,2,8,3,3,5,5,4,2,2,2,4,4,6,6, 
	 2,5,2,8,4,4,6,6,2,4,2,7,5,5,7,7, 
	 6,6,2,8,3,3,5,5,3,2,2,2,3,4,6,6, 
	 2,5,2,8,4,4,6,6,2,4,2,7,5,5,7,7, 
	 6,6,2,8,3,3,5,5,4,2,2,2,5,4,6,6, 
	 2,5,2,8,4,4,6,6,2,4,2,7,5,5,7,7, 
	 2,6,2,6,3,3,3,3,2,2,2,2,4,4,4,4, 
	 2,6,2,6,4,4,4,4,2,5,2,5,5,5,5,5, 
	 2,6,2,6,3,3,3,3,2,2,2,2,4,4,4,4, 
	 2,5,2,5,4,4,4,4,2,4,2,5,4,4,4,4, 
	 2,6,2,8,3,3,5,5,2,2,2,2,4,4,6,6, 
	 2,5,2,8,4,4,6,6,2,4,2,7,5,5,7,7, 
	 2,6,2,8,3,3,5,5,2,2,2,2,4,4,6,6, 
	 2,5,2,8,4,4,6,6,2,4,2,7,5,5,7,7 
	}; 
 

	private int ZNTable[] = 
	{ 
	  Z_FLAG,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
	  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
	  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
	  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
	  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
	  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
	  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
	  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
	  N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG, 
	  N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG, 
	  N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG, 
	  N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG, 
	  N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG, 
	  N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG, 
	  N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG, 
	  N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG, 
	  N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG, 
	  N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG, 
	  N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG, 
	  N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG, 
	  N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG, 
	  N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG, 
	  N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG, 
	  N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG,N_FLAG, 
	}; 
 

	/** Memory */ 
	public final int  mem[] = new int[ 65536 ];





	/** Memory access - rewrite in subclasses */
        public int peek( int addr ) {
		return mem[ addr ];
	}

	public void poke( int addr, int newByte ) {
		mem[ addr ] = newByte;
	}

	/** Program access - not used yet */
	public int OpCodePeek( int addr ) {
		return peek( addr );
	}


	/** Word access to memory (to read address) */
	public int peekw( int addr ) {
		addr &= 0xffff;
		return ((peek(addr+1)<<8) + peek(addr));

	}


        /** Periodic handler - rewrite in subclasses - called before each 
                               instruction */
        public void preExecution (int cycles) {
	}


	/** Periodic handler - rewrite in subclasses and return interrupt 
	                       type when hardware interrupt occurs */
	public int Loop6502() {
		// return INT_NONE, INT_NMI, INT_IRQ, or INT_QUIT in your subclass
		return INT_NONE;
	}


	/** Stack access */
	public final void push( int PByte ) {
		poke((0x0100|S), PByte);
		S = (255+S)&0x00ff;
	} 

	public final int pop() {
                S = (S+1)&0x00ff;
                return(peek(0x0100|S));
	} 

	public final void popPC() {
		PC =  pop();      //Lo
		PC += (pop()<<8); //Hi
	} 

	public final void pushPC() {
		push(PC>>8);     //Hi
		push(PC&0x00ff); //Lo
	} 


	/** Reset all registers to power on state */
	public final void reset() {
		A  = X = Y = 0x00;
		P  = (Z_FLAG | R_FLAG);
		S  = 0xff;
		PC = peekw(0xFFFC); // reset vector
		ICount   = IPeriod;
		IRequest = INT_NONE;
		AfterCLI = false;

	}


	/** Perform an interrupt */
	public void interrupt(int Type) {
		if((Type==INT_NMI) || ((Type==INT_IRQ) && ((P&I_FLAG)==0))) { 
			ICount -= 7; 
    			pushPC(); 
			push(P&~B_FLAG); 
			P &= ~D_FLAG; 
			if (Type==INT_NMI)  PC=peekw(0xFFFA);   // NMI
			else { P |= I_FLAG; PC=peekw(0xFFFE); } // IRQ
		} 
	}


	/** 6502 fetch/execute loop */
	public final void execute() {
		int I,Instr;

		while ( true ) {

		if (Debugging) printDebug();
		//try { int bbb=System.in.read();} catch (Exception e) {};

    		Instr  =  OpCodePeek(PC++);
                ICount -= Cycles[Instr];

                preExecution(Cycles[Instr]);    // call periodic handler

		switch(Instr) {

		case 0x10: /* BPL * REL */ { branchFlagClr(N_FLAG); break; }
		case 0x30: /* BMI * REL */ { branchFlagSet(N_FLAG); break; }
		case 0xD0: /* BNE * REL */ { branchFlagClr(Z_FLAG); break; }
		case 0xF0: /* BEQ * REL */ { branchFlagSet(Z_FLAG); break; }
		case 0x90: /* BCC * REL */ { branchFlagClr(C_FLAG); break; }
		case 0xB0: /* BCS * REL */ { branchFlagSet(C_FLAG); break; }
		case 0x50: /* BVC * REL */ { branchFlagClr(V_FLAG); break; }
		case 0x70: /* BVS * REL */ { branchFlagSet(V_FLAG); break; }

	
		case 0x40: /* RTI */ {
			P = (pop()|R_FLAG);
			popPC();
			break; 
 		}

		case 0x60: /* RTS */ {
			popPC(); 
			PC++;
			break; 
		}

		case 0x20: /* JSR $ssss ABS */ {
			I = PC; PC++;
			pushPC();
			PC = peekw(I);
			break; 
		}
		
  
		case 0x4C: /* JMP $ssss ABS */ {
			PC = peekw(PC);
			break; 
		}

		case 0x6C: /* JMP ($ssss) ABDINDIR */ {
			PC = peekw(peekw(PC));
			break; 
		}


		case 0x00: /* BRK */ {
			PC++;
			pushPC();
			push(P|B_FLAG);
			P  = (P|I_FLAG)&~D_FLAG;
			PC = peekw(0xFFFE);
			break; 
		}
		
		case 0x58: /* CLI */ {
 			if ((IRequest != INT_NONE) && ((P&I_FLAG) != 0)) { 
 				AfterCLI = true; 
				IBackup  = ICount; 
				ICount   = 1; 
			} 
			P &= ~I_FLAG; 
			break;
		}

		case 0x28: /* PLP */ { 
			I = pop();
			if ((IRequest != INT_NONE) && (((I^P)&~I&I_FLAG)!=0) ) { 
				AfterCLI = true; 
				IBackup  = ICount; 
				ICount   = 1; 
			} 
			P = (I|R_FLAG); 
			break; 
		}

		case 0x08: /* PHP */ { push(P); break; }
		case 0x18: /* CLC */ { P &= ~C_FLAG; break; }
		case 0xB8: /* CLV */ { P &= ~V_FLAG; break; }
		case 0xD8: /* CLD */ { P &= ~D_FLAG; break; }
		case 0x38: /* SEC */ { P |=  C_FLAG; break; }
		case 0xF8: /* SED */ { P |=  D_FLAG; break; }
		case 0x78: /* SEI */ { P |=  I_FLAG; break; }
		case 0x48: /* PHA */ { push(A); break; } 
		case 0x68: /* PLA */ { A = pop(); 	calcFlags(A); break; }
		case 0x98: /* TYA */ { A = Y;  		calcFlags(A); break; }
		case 0xA8: /* TAY */ { Y = A;  		calcFlags(Y); break; }
		case 0xC8: /* INY */ { Y++; Y &= 0xff; 	calcFlags(Y); break; }
		case 0x88: /* DEY */ { Y =(Y+255)&0xff;	calcFlags(Y); break; }
		case 0x8A: /* TXA */ { A = X;  		calcFlags(A); break; }
		case 0xAA: /* TAX */ { X = A;  		calcFlags(X); break; }
		case 0xE8: /* INX */ { X++; X &= 0xff; 	calcFlags(X); break; }
		case 0xCA: /* DEX */ { X =(X+255)&0xff;	calcFlags(X); break; }
		case 0xEA: /* NOP */ { break; }
		case 0x9A: /* TXS */ { S = X; break; }
		case 0xBA: /* TSX */ { X = S; break; }

		case 0x24: /* BIT $ss ZP */ {
			I = rd_Zp();
			P &= ~(N_FLAG|V_FLAG|Z_FLAG); 
			P |= ((I&(N_FLAG|V_FLAG)) | (((I&A)!=0) ? 0:Z_FLAG));
			PC++;
			break;
		}

		case 0x2C: /* BIT $ssss ABS */ {
			I = rd_Ab();
			P &= ~(N_FLAG|V_FLAG|Z_FLAG); 
			P |= ((I&(N_FLAG|V_FLAG)) | (((I&A)!=0) ? 0:Z_FLAG)); 
			PC+=2;
			break;
		}



		// *** Zero page addressing

		case 0x05: /* ORA $ss ZP */ {
			I = rd_Zp();
			ora(I);
			PC++;
			break;
		}

		case 0x06: /* ASL $ss ZP */ {
			I = rd_Zp();
			wr_back( asl(I) );
			PC++;
			break;
		}

		case 0x25: /* AND $ss ZP */ {
			I = rd_Zp();
			and(I);
			PC++;
			break;
		}

		case 0x26: /* ROL $ss ZP */ {
			I = rd_Zp();
			wr_back( rol(I) );
			PC++;
			break;
		}

		case 0x45: /* EOR $ss ZP */ {
			I = rd_Zp();
			eor(I);
			PC++;
			break;
		}

		case 0x46: /* LSR $ss ZP */ {
			I = rd_Zp();
			wr_back( lsr(I) );
			PC++;
			break;
		}

		case 0x65: /* ADC $ss ZP */ {
			I = rd_Zp();
			adc(I);
			PC++;
			break;
		}

		case 0x66: /* ROR $ss ZP */ {
			I = rd_Zp();
			wr_back( ror(I) );
			PC++;
			break;
		}

		case 0x84: /* STY $ss ZP */ { wr_Zp(Y); PC++; break; }
		case 0x85: /* STA $ss ZP */ { wr_Zp(A); PC++; break; }
		case 0x86: /* STX $ss ZP */ { wr_Zp(X); PC++; break; }
		case 0xA4: /* LDY $ss ZP */ { Y = rd_Zp(); calcFlags(Y); PC++; break; }
		case 0xA5: /* LDA $ss ZP */ { A = rd_Zp(); calcFlags(A); PC++; break; }
		case 0xA6: /* LDX $ss ZP */ { X = rd_Zp(); calcFlags(X); PC++; break; }

		case 0xC4: /* CPY $ss ZP */ {
			I = rd_Zp();
                        cmp(Y, I);
			PC++;
			break;
		}

		case 0xC5: /* CMP $ss ZP */ {
			I = rd_Zp();
                        cmp(A, I);
			PC++;
			break;
		}

		case 0xC6: /* DEC $ss ZP */ {
			I = rd_Zp();
			wr_back( dec(I) );
			PC++;
			break;
		}

		case 0xE4: /* CPX $ss ZP */ {
			I = rd_Zp();
                        cmp(X, I);
			PC++;
			break;
		}

		case 0xE5: /* SBC $ss ZP */ {
			I = rd_Zp();
			sbc(I);
			PC++;
			break;
		}

		case 0xE6: /* INC $ss ZP */ {
			I = rd_Zp();
			wr_back( inc(I) );
			PC++;
			break;
		}



 		// *** Absolute addressing

		case 0x0D: /* ORA $ssss ABS */ {
			I = rd_Ab();
			ora(I);
			PC+=2;
			break;
		}

                case 0x0E: /* ASL $ssss ABS */ {
			I = rd_Ab();
			wr_back( asl(I) );
			PC+=2;
			break;
		}

                case 0x2D: /* AND $ssss ABS */ {
                        I = rd_Ab();
			and(I);
			PC+=2;
			break;
		}

                case 0x2E: /* ROL $ssss ABS */ {
                        I = rd_Ab();
			wr_back( rol(I) );
			PC+=2;
			break;
		}

                case 0x4D: /* EOR $ssss ABS */ {
                        I = rd_Ab();
			eor(I);
			PC+=2;
			break;
		}

                case 0x4E: /* LSR $ssss ABS */ {
                        I = rd_Ab();
			wr_back( lsr(I) );
			PC+=2;
			break;
		}

                case 0x6D: /* ADC $ssss ABS */ {
                        I = rd_Ab();
			adc(I);
			PC+=2;
			break;
		}

                case 0x6E: /* ROR $ssss ABS */ {
                        I = rd_Ab();
			wr_back( ror(I) );
			PC+=2;
			break;
		}

                case 0x8C: /* STY $ssss ABS */ { wr_Ab(Y); PC+=2; break; }
                case 0x8D: /* STA $ssss ABS */ { wr_Ab(A); PC+=2; break; }
                case 0x8E: /* STX $ssss ABS */ { wr_Ab(X); PC+=2; break; }
                case 0xAC: /* LDY $ssss ABS */ { Y = rd_Ab(); calcFlags(Y); PC+=2; break; }
                case 0xAD: /* LDA $ssss ABS */ { A = rd_Ab(); calcFlags(A); PC+=2; break; }
                case 0xAE: /* LDX $ssss ABS */ { X = rd_Ab(); calcFlags(X); PC+=2; break; }

                case 0xCC: /* CPY $ssss ABS */ {
                        I = rd_Ab();
                        cmp(Y, I);
			PC+=2;
			break;
		}

                case 0xCD: /* CMP $ssss ABS */ {
                        I = rd_Ab();
                        cmp(A, I);
			PC+=2;
			break;
		}

                case 0xCE: /* DEC $ssss ABS */ {
                        I = rd_Ab();
			wr_back( dec(I) );
			PC+=2;
			break;
		}

                case 0xEC: /* CPX $ssss ABS */ {
                        I = rd_Ab();
                        cmp(X, I);
			PC+=2;
			break;
		}

                case 0xED: /* SBC $ssss ABS */ {
                        I = rd_Ab();
			sbc(I);
			PC+=2;
			break;
		}

                case 0xEE: /* INC $ssss ABS */ {
                        I = rd_Ab();
			wr_back( inc(I) );
			PC+=2;
			break;
		}


 		// *** Immediate addressing

                case 0x09: /* ORA #$ss IMM */ {
                        I = rd_Im();
			ora(I);
			PC++;
			break;
		}

                case 0x29: /* AND #$ss IMM */ {
                        I = rd_Im();
			and(I);
			PC++;
			break;
		}

                case 0x49: /* EOR #$ss IMM */ {
                        I = rd_Im();
			eor(I);
			PC++;
			break;
		}

                case 0x69: /* ADC #$ss IMM */ {
                        I = rd_Im();
			adc(I);
			PC++;
			break;
		}

                case 0xA0: /* LDY #$ss IMM */ { Y = rd_Im(); calcFlags(Y); PC++; break; }
                case 0xA2: /* LDX #$ss IMM */ { X = rd_Im(); calcFlags(X); PC++; break; }
                case 0xA9: /* LDA #$ss IMM */ { A = rd_Im(); calcFlags(A); PC++; break; }

                case 0xC0: /* CPY #$ss IMM */ {
                        I = rd_Im();
                        cmp(Y, I);
			PC++;
			break;
		}

                case 0xC9: /* CMP #$ss IMM */ {
                        I = rd_Im();
                        cmp(A, I);
			PC++;
			break;
		}

                case 0xE0: /* CPX #$ss IMM */ {
                        I = rd_Im();
                        cmp(X, I);
			PC++;
			break;
		}

                case 0xE9: /* SBC #$ss IMM */ {
                        I = rd_Im();
			sbc(I);
			PC++;
			break;
		}



 		// *** Zero page X addressing

                case 0x15: /* ORA $ss,x ZP,x */ {
                        I = rd_Zx();
			ora(I);
			PC++;
			break;
		}

                case 0x16: /* ASL $ss,x ZP,x */ {
                        I = rd_Zx();
			wr_back( asl(I) );
			PC++;
			break;
		}

                case 0x35: /* AND $ss,x ZP,x */ {
                        I = rd_Zx();
			and(I);
			PC++;
			break;
		}

                case 0x36: /* ROL $ss,x ZP,x */ {
                        I = rd_Zx();
			wr_back( rol(I) );
			PC++;
			break;
		}

                case 0x55: /* EOR $ss,x ZP,x */ {
                        I = rd_Zx();
			eor(I);
			PC++;
			break;
		}

                case 0x56: /* LSR $ss,x ZP,x */ {
                        I = rd_Zx();
			wr_back( lsr(I) );
			PC++;
			break;
		}

                case 0x75: /* ADC $ss,x ZP,x */ {
                        I = rd_Zx();
			adc(I);
			PC++;
			break;
		}

                case 0x76: /* ROR $ss,x ZP,x */ {
                        I = rd_Zx();
			wr_back( ror(I) );
			PC++;
			break;
		}

                case 0x94: /* STY $ss,x ZP,x */ { wr_Zx(Y); PC++; break; }
                case 0x95: /* STA $ss,x ZP,x */ { wr_Zx(A); PC++; break; }
                case 0x96: /* STX $ss,y ZP,y */ { wr_Zy(X); PC++; break; }
                case 0xB4: /* LDY $ss,x ZP,x */ { Y = rd_Zx(); calcFlags(Y); PC++; break; }
                case 0xB5: /* LDA $ss,x ZP,x */ { A = rd_Zx(); calcFlags(A); PC++; break; }
                case 0xB6: /* LDX $ss,y ZP,y */ { X = rd_Zy(); calcFlags(X); PC++; break; }

                case 0xD5: /* CMP $ss,x ZP,x */ {
                        I = rd_Zx();
			cmp(A, I);
			PC++;
			break;
		}

                case 0xD6: /* DEC $ss,x ZP,x */ {
                        I = rd_Zx();
			wr_back( dec(I) );
			PC++;
			break;
		}

                case 0xF5: /* SBC $ss,x ZP,x */ {
                        I = rd_Zx();
			sbc(I);
			PC++;
			break;
		}

                case 0xF6: /* INC $ss,x ZP,x */ {
                        I = rd_Zx();
			wr_back( inc(I) );
			PC++;
			break;
		}

  

 		// *** Absolute X and Y addressing

                case 0x19: /* ORA $ssss,y ABS,y */ {
                        I = rd_Ay();
			ora(I);
			PC+=2;
			break;
		}

                case 0x1D: /* ORA $ssss,x ABS,x */ {
                        I = rd_Ax();
			ora(I);
			PC+=2;
			break;
		}

                case 0x1E: /* ASL $ssss,x ABS,x */ {
                        I = rd_Ax();
			wr_back( asl(I) );
			PC+=2;
			break;
		}

                case 0x39: /* AND $ssss,y ABS,y */ {
                        I = rd_Ay();
			and(I);
			PC+=2;
			break;
		}

		case 0x3D: /* AND $ssss,x ABS,x */ {
                        I = rd_Ax();
			and(I);
			PC+=2;
			break;
		}

		case 0x3E: /* ROL $ssss,x ABS,x */ {
                        I = rd_Ax();
			wr_back( rol(I) );
			PC+=2;
			break;
		}

                case 0x59: /* EOR $ssss,y ABS,y */ {
                        I = rd_Ay();
			eor(I);
			PC+=2;
			break;
		}

                case 0x5D: /* EOR $ssss,x ABS,x */ {
                        I = rd_Ax();
			eor(I);
			PC+=2;
			break;
		}

                case 0x5E: /* LSR $ssss,x ABS,x */ {
                        I = rd_Ax();
			wr_back( lsr(I) );
			PC+=2;
			break;
		}

                case 0x79: /* ADC $ssss,y ABS,y */ {
                        I = rd_Ay();
			adc(I);
			PC+=2;
			break;
		}

                case 0x7D: /* ADC $ssss,x ABS,x */ {
                        I = rd_Ax();
			adc(I);
			PC+=2;
			break;
		}

                case 0x7E: /* ROR $ssss,x ABS,x */ {
                        I = rd_Ax();
			wr_back( ror(I) );
			PC+=2;
			break;
		}

                case 0x99: /* STA $ssss,y ABS,y */ { wr_Ay(A); PC+=2; break; }
                case 0x9D: /* STA $ssss,x ABS,x */ { wr_Ax(A); PC+=2; break; }
                case 0xB9: /* LDA $ssss,y ABS,y */ { A = rd_Ay(); calcFlags(A); PC+=2; break; }
                case 0xBC: /* LDY $ssss,x ABS,x */ { Y = rd_Ax(); calcFlags(Y); PC+=2; break; }
                case 0xBD: /* LDA $ssss,x ABS,x */ { A = rd_Ax(); calcFlags(A); PC+=2; break; }
                case 0xBE: /* LDX $ssss,y ABS,y */ { X = rd_Ay(); calcFlags(X); PC+=2; break; }

                case 0xD9: /* CMP $ssss,y ABS,y */ {
                        I = rd_Ay();
                        cmp(A, I);
			PC+=2;
			break;
		}

                case 0xDD: /* CMP $ssss,x ABS,x */ {
                        I = rd_Ax();
                        cmp(A, I);
			PC+=2;
			break;
		}

                case 0xDE: /* DEC $ssss,x ABS,x */ {
                        I = rd_Ax();
			wr_back( dec(I) );
			PC+=2;
			break;
		}

                case 0xF9: /* SBC $ssss,y ABS,y */ {
                        I = rd_Ay();
			sbc(I);
			PC+=2;
			break;
		}

                case 0xFD: /* SBC $ssss,x ABS,x */ {
                        I = rd_Ax();
			sbc(I);
			PC+=2;
			break;
		}

                case 0xFE: /* INC $ssss,x ABS,x */ {
                        I = rd_Ax();
			wr_back( inc(I) );
			PC+=2;
			break;
		}



 		// *** Indirect index addressing

                case 0x01: /* ORA ($ss,x) INDEXINDIR */ {
                        I = rd_Ix();
			ora(I);
			PC++;
			break;
		}

                case 0x11: /* ORA ($ss),y INDIRINDEX */ {
                        I = rd_Iy();
			ora(I);
			PC++;
			break;
		}

                case 0x21: /* AND ($ss,x) INDEXINDIR */ {
                        I = rd_Ix();
			and(I);
			PC++;
			break;
		}

                case 0x31: /* AND ($ss),y INDIRINDEX */ {
                        I = rd_Iy();
			and(I);
			PC++;
			break;
		}

                case 0x41: /* EOR ($ss,x) INDEXINDIR */ {
                        I = rd_Ix();
			eor(I);
			PC++;
			break;
		}

                case 0x51: /* EOR ($ss),y INDIRINDEX */ {
                        I = rd_Iy();
			eor(I);
			PC++;
			break;
		}

                case 0x61: /* ADC ($ss,x) INDEXINDIR */ {
                        I = rd_Ix();
			adc(I);
			PC++;
			break;
		}

                case 0x71: /* ADC ($ss),y INDIRINDEX */ {
                        I = rd_Iy();
			adc(I);
			PC++;
			break;
		}

                case 0x81: /* STA ($ss,x) INDEXINDIR */ { wr_Ix(A); PC++; break; }
                case 0x91: /* STA ($ss),y INDIRINDEX */ { wr_Iy(A); PC++; break; }
                case 0xA1: /* LDA ($ss,x) INDEXINDIR */ { A = rd_Ix(); calcFlags(A); PC++; break; }
                case 0xB1: /* LDA ($ss),y INDIRINDEX */ { A = rd_Iy(); calcFlags(A); PC++; break; }

                case 0xC1: /* CMP ($ss,x) INDEXINDIR */ {
                        I = rd_Ix();
			cmp(A, I);
			PC++;
			break;
		}

                case 0xD1: /* CMP ($ss),y INDIRINDEX */ {
                        I = rd_Iy();
			cmp(A, I);
			PC++;
			break;
		}


                case 0xE1: /* SBC ($ss,x) INDEXINDIR */ {
                        I = rd_Ix();
			sbc(I);
			PC++;
			break;
		}

                case 0xF1: /* SBC ($ss),y INDIRINDEX */ {
                        I = rd_Iy();
			sbc(I);
			PC++;
			break;
		}


		// *** Accumulator addressing
		case 0x0A: /* ASL a ACC */ { A = asl(A); break; }
		case 0x2A: /* ROL a ACC */ { A = rol(A); break; }
		case 0x4A: /* LSR a ACC */ { A = lsr(A); break; }
		case 0x6A: /* ROR a ACC */ { A = ror(A); break; }
 

		default: {
			System.out.println(
				"[J6502] Unrecognized instruction: "+Instr+
				" at PC="+(PC-1)+
				"--"+mem[PC]+
				"--"+mem[PC+1]+
				"--"+mem[PC+2]+
				"--"+mem[PC+3]
			);
 			break;
		}


		} // end switch


    		/* If cycle counter expired... */ 
		if(ICount<=0) { 
		      /* If we have come after CLI, get INT_? from IRequest */ 
		      /* Otherwise, get it from the loop handler            */ 
			if (AfterCLI) { 
		       		I = IRequest; 		/* Get pending interrupt     */ 
				ICount += IBackup-1;	/* Restore the ICount        */ 
				AfterCLI = false;	/* Done with AfterCLI state  */ 
	      		} 
      			else { 
        			I = Loop6502();		/* Call the periodic handler */ 
        			ICount = IPeriod;	/* Reset the cycle counter   */ 
	      		} 
 
      			if (I==INT_QUIT) return;	/* Exit if INT_QUIT     */ 
      			if (I!=INT_NONE) interrupt(I);	/* Interrupt if needed  */  
    		} 

 		} // end while 
 
	} 




 
	/** Reading From Memory **************************************/ 
	/** These methods calculate the address and read from it.   **/ 
	/*************************************************************/ 

	private final int rd_Ab () { LastRdAddr = peekw(PC); 			return peek(LastRdAddr); }
	private final int rd_Im () { return peek(PC); }
	private final int rd_Zp () { LastRdAddr = peek(PC); 			return peek(LastRdAddr); }
	private final int rd_Zx () { LastRdAddr = (peek(PC)+X)&0xff; 		return peek(LastRdAddr); }
	private final int rd_Zy () { LastRdAddr = (peek(PC)+Y)&0xff; 		return peek(LastRdAddr); }
	private final int rd_Ax () { LastRdAddr = (peekw(PC)+X)&0xffff; 	return peek(LastRdAddr); }
	private final int rd_Ay () { LastRdAddr = (peekw(PC)+Y)&0xffff; 	return peek(LastRdAddr); }
	private final int rd_Ix () { LastRdAddr = peekw((peek(PC)+X)&0xff);   return peek(LastRdAddr); }
	private final int rd_Iy () { LastRdAddr = (peekw(peek(PC))+Y)&0xffff;	return peek(LastRdAddr); }


	/** Writing To Memory ****************************************/ 
	/** These methods calculate the address and write to it.    **/ 
	/*************************************************************/ 

	private final void wr_Ab (int v) { poke(peekw(PC), v); }
	private final void wr_Zp (int v) { poke(peek(PC), v); }
	private final void wr_Zx (int v) { poke((peek(PC)+X)&0xff, v); }
	private final void wr_Zy (int v) { poke((peek(PC)+Y)&0xff, v); }
	private final void wr_Ax (int v) { poke((peekw(PC)+X)&0xffff, v); }
	private final void wr_Ay (int v) { poke((peekw(PC)+Y)&0xffff, v); }
	private final void wr_Ix (int v) { poke( peekw( (peek(PC)+X)&0xff ), v); }
	private final void wr_Iy (int v) { poke( (peekw(peek(PC))+Y)&0xffff, v); }

	private final void wr_back (int v) { poke(LastRdAddr, v); }
 
	/** Other Methods ********************************************/ 
	/** Calculating flags, stack, jumps, arithmetics, etc.      **/ 
	/*************************************************************/ 

	/** Calculate flags based on register Rg */
	private final void calcFlags (int Rg) {
		P = (P & ~(Z_FLAG|N_FLAG)) | ZNTable[Rg];
	}

	/** Branch on flag set */
	private final void branchFlagSet(int flag) {
		int b;
		if ((P&flag)==0) PC++;
		else {
			b=peek(PC);
			if (b>0x7f) b-=256;
			PC+=(b+1);
			ICount--;
		}		
	}

	/** Branch on flag clear */
	private final void branchFlagClr(int flag) {
		int b;
		if ((P&flag)!=0) PC++;
		else {
			b=peek(PC);
			if (b>0x7f) b=-(256-b);
			PC+=(b+1);
			ICount--;
		}		
	}



	/** Add with carry */
	private final void adc(int Rg) {
		int z, zl, zh;
		if ((P&D_FLAG)==0) {   // no decimal mode 
			z = A+Rg+(P&C_FLAG);
			zl = z & 0xff;
                        zh = (z >> 8)&0xff;
			P &= ~(N_FLAG|V_FLAG|Z_FLAG|C_FLAG);
			P |= 	(((~(A^Rg)&(A^zl)&0x80) != 0) ? V_FLAG:0) | 
				((zh != 0) ? C_FLAG:0) |ZNTable[zl];
			A = zl;
		}
		else {   // decimal mode 
                        zl = ((A&0x0F)+(Rg&0x0F)+(P&C_FLAG))&0xff;
                        zh = ((A>>4)+(Rg>>4)+(zl>15? 1:0))&0xff;
			if(zl>9) { zl += 6; zh++; }
			if(zh>9)   zh += 6; 
                        A = (zl&0x0F)|((zh<<4)&0xff);
			P = (P&~C_FLAG)|(zh>15? C_FLAG:0); 
		}		
	}


	/** Subtract with carry */
        private final void sbc(int Rg) {
                int z, zl, zh;
                if ((P&D_FLAG)==0) {   // no decimal mode
                        z = A-Rg-(~P&C_FLAG);
                        zl = z & 0xff;
                        zh = z >> 8;
                        P &= ~(N_FLAG|V_FLAG|Z_FLAG|C_FLAG);
                        P |=    ((((A^Rg)&(A^zl)&0x80) != 0) ? V_FLAG:0)|
                                ((zh != 0)? 0:C_FLAG)|ZNTable[zl];
                        A = zl;
                }
                else {   // decimal mode
                        zl = ((A&0x0F)-(Rg&0x0F)+(~P&C_FLAG))&0xff;
                        if(zl>15) { zl -= 6; }
                        zh = (A>>4)-(Rg>>4)-(zl&0x10);
                        if(zh>15)   zh -= 6;
                        A = (zl&0x0F)|((zh<<4)&0xff);
                        P = (P&~C_FLAG)|(zh>15? C_FLAG:0);
                }
        }







	/** Compare */
	private final void cmp(int Rg1, int Rg2) {
		int z;
		z = Rg1-Rg2;
		P &= ~(N_FLAG|Z_FLAG|C_FLAG);
		P |= ZNTable[z&0xff]|(((z>>8)!=0)? 0:C_FLAG); 
	}


	/** Logical operators, increment, and decrement */
	private final void and(int Rg) { A &= Rg; calcFlags(A); }
	private final void ora(int Rg) { A |= Rg; calcFlags(A); }
	private final void eor(int Rg) { A ^= Rg; calcFlags(A); }
	private final int  inc(int Rg) { Rg++; Rg&=0xff;    	calcFlags(Rg); return(Rg); }
	private final int  dec(int Rg) { Rg=(Rg+255)&0xff; calcFlags(Rg); return(Rg); }


	/** Shift and rotate */
	private final int  asl(int Rg) { P&=~C_FLAG; P|=Rg>>7; Rg<<=1; Rg&=0xff; calcFlags(Rg); return(Rg); }
	private final int  lsr(int Rg) { P&=~C_FLAG; P|=Rg&C_FLAG; Rg>>=1; Rg&=0xff; calcFlags(Rg); return(Rg); }
	private final int rol(int Rg) { 
		int z;
		z = ((Rg<<1)|(P&C_FLAG))&0xff;
		P &= ~C_FLAG; P |= Rg>>7;
		Rg = z;
		calcFlags(Rg);
		return(Rg);
	}
	private final int ror(int Rg) { 
		int z;
		z = ((Rg>>1)|(P<<7))&0xff;
		P &= ~C_FLAG; P |= Rg&C_FLAG;
		Rg = z;
		calcFlags(Rg);
		return(Rg);
	}




	/****************** 6502 Disassembler  ********************/


	/* Addressing methods used by the 6502 CPU:   */ 
 
	private int Ac = 0;
	private int Il = 1;
	private int Im = 2;
	private int Ab = 3;
	private int Zp = 4;
	private int Zx = 5;
	private int Zy = 6;
	private int Ax = 7;
	private int Ay = 8;
	private int Rl = 9;
	private int Ix = 10;
	private int Iy = 11;
	private int In = 12;
	private int No = 13;
  
	private String MN[] = 
		{ 
		"adc","and","asl","bcc","bcs","beq","bit","bmi", 
		"bne","bpl","brk","bvc","bvs","clc","cld","cli", 
		"clv","cmp","cpx","cpy","dec","dex","dey","inx", 
		"iny","eor","inc","jmp","jsr","lda","nop","ldx", 
		"ldy","lsr","ora","pha","php","pla","plp","rol", 
		"ror","rti","rts","sbc","sta","stx","sty","sec", 
		"sed","sei","tax","tay","txa","tya","tsx","txs" 
		}; 
 
	private int AD[] = 
		{ 
		  10,Il, 34,Ix, No,No, No,No, No,No, 34,Zp,  2,Zp, No,No, 
		  36,Il, 34,Im,  2,Ac, No,No, No,No, 34,Ab,  2,Ab, No,No, 
		   9,Rl, 34,Iy, No,No, No,No, No,No, 34,Zx,  2,Zx, No,No, 
		  13,Il, 34,Ay, No,No, No,No, No,No, 34,Ax,  2,Ax, No,No, 
		  28,Ab,  1,Ix, No,No, No,No,  6,Zp,  1,Zp, 39,Zp, No,No, 
		  38,Il,  1,Im, 39,Ac, No,No,  6,Ab,  1,Ab, 39,Ab, No,No, 
		   7,Rl,  1,Iy, No,No, No,No, No,No,  1,Zx, 39,Zx, No,No, 
		  47,Il,  1,Ay, No,No, No,No, No,No,  1,Ax, 39,Ax, No,No, 
		  41,Il, 25,Ix, No,No, No,No, No,No, 25,Zp, 33,Zp, No,No, 
		  35,Il, 25,Im, 33,Ac, No,No, 27,Ab, 25,Ab, 33,Ab, No,No, 
		  11,Rl, 25,Iy, No,No, No,No, No,No, 25,Zx, 33,Zx, No,No, 
		  15,Il, 25,Ay, No,No, No,No, No,No, 25,Ax, 33,Ax, No,No, 
		  42,Il,  0,Ix, No,No, No,No, No,No,  0,Zp, 40,Zp, No,No, 
		  37,Il,  0,Im, 40,Ac, No,No, 27,In,  0,Ab, 40,Ab, No,No, 
		  12,Rl,  0,Iy, No,No, No,No, No,No,  0,Zx, 40,Zx, No,No, 
		  49,Il,  0,Ay, No,No, No,No, No,No,  0,Ax, 40,Ax, No,No, 
		  No,No, 44,Ix, No,No, No,No, 46,Zp, 44,Zp, 45,Zp, No,No, 
		  22,Il, No,No, 52,Il, No,No, 46,Ab, 44,Ab, 45,Ab, No,No, 
		   3,Rl, 44,Iy, No,No, No,No, 46,Zx, 44,Zx, 45,Zy, No,No, 
		  53,Il, 44,Ay, 55,Il, No,No, No,No, 44,Ax, No,No, No,No, 
		  32,Im, 29,Ix, 31,Im, No,No, 32,Zp, 29,Zp, 31,Zp, No,No, 
		  51,Il, 29,Im, 50,Il, No,No, 32,Ab, 29,Ab, 31,Ab, No,No, 
		   4,Rl, 29,Iy, No,No, No,No, 32,Zx, 29,Zx, 31,Zy, No,No, 
		  16,Il, 29,Ay, 54,Il, No,No, 32,Ax, 29,Ax, 31,Ay, No,No, 
		  19,Im, 17,Ix, No,No, No,No, 19,Zp, 17,Zp, 20,Zp, No,No, 
		  24,Il, 17,Im, 21,Il, No,No, 19,Ab, 17,Ab, 20,Ab, No,No, 
		   8,Rl, 17,Iy, No,No, No,No, No,No, 17,Zx, 20,Zx, No,No, 
		  14,Il, 17,Ay, No,No, No,No, No,No, 17,Ax, 20,Ax, No,No, 
		  18,Im, 43,Ix, No,No, No,No, 18,Zp, 43,Zp, 26,Zp, No,No, 
		  23,Il, 43,Im, 30,Il, No,No, 18,Ab, 43,Ab, 26,Ab, No,No, 
		   5,Rl, 43,Iy, No,No, No,No, No,No, 43,Zx, 26,Zx, No,No, 
		  48,Il, 43,Ay, No,No, No,No, No,No, 43,Ax, 26,Ax, No,No 
		}; 
 

	/** DAsm(PC) **************************************************/         
	/** This function will disassemble a single command and      **/     
	/** return a string with the disassembled command.           **/   
	/**************************************************************/ 
	public String DAsm(int ProgramCounter) {

		int PrgCnt = ProgramCounter;
		int OP = mem[PrgCnt++]*2;
		int Z,J;
		String DA;
		
		switch(AD[OP+1]) { 
			case 0: /*Ac*/ { DA = MN[AD[OP]]+" a"; break; } 
			case 1: /*Il*/ { DA = MN[AD[OP]]; break; }
 
			case 9: /*Rl*/ {
				J = mem[PrgCnt++];
				Z = ProgramCounter+2+((J<0x80)? J:(J-256));  
				DA = MN[AD[OP]]+ " $"+HexStr(Z>>8)+HexStr(Z&0xff);
				break; }
 
			case 2: /*Im*/ { DA = MN[AD[OP]]+" #$"+HexStr(mem[PrgCnt++]); break; } 
			case 4: /*Zp*/ { DA = MN[AD[OP]]+" $" +HexStr(mem[PrgCnt++]); break; }
			case 5: /*Zx*/ { DA = MN[AD[OP]]+" $" +HexStr(mem[PrgCnt++])+",x"; break; }
			case 6: /*Zy*/ { DA = MN[AD[OP]]+" $" +HexStr(mem[PrgCnt++])+",y"; break; }
			case 10:/*Ix*/ { DA = MN[AD[OP]]+" ($"+HexStr(mem[PrgCnt++])+",x)"; break; }
			case 11:/*Iy*/ { DA = MN[AD[OP]]+" ($"+HexStr(mem[PrgCnt++])+"),y"; break; }
 
			case 3: /*Ab*/ { DA = MN[AD[OP]]+" $"	+HexStr(mem[PrgCnt+1])
													+HexStr(mem[PrgCnt]);
					   PrgCnt+=2;
					   break; }
			case 7: /*Ax*/ { DA = MN[AD[OP]]+" $"	+HexStr(mem[PrgCnt+1])
													+HexStr(mem[PrgCnt])+",x";
					   PrgCnt+=2;
					   break; }
			case 8: /*Ay*/ { DA = MN[AD[OP]]+" $"	+HexStr(mem[PrgCnt+1])
													+HexStr(mem[PrgCnt])+",y";
					   PrgCnt+=2;
					   break; }
			case 12:/*In*/ { DA = MN[AD[OP]]+" ($"	+HexStr(mem[PrgCnt+1])
													+HexStr(mem[PrgCnt])+")";
					   PrgCnt+=2;
					   break; }
			default: { DA = ".db $"+HexStr((int)(OP/2))+" <Invalid OPcode>"; break; }

			} 
	return DA;

	}


	/* Print byte in Hexadecimal notation */
	public String HexStr(int B) {
		String HexCd[] = {	"0","1","2","3","4","5","6","7",
							"8","9","a","b","c","d","e","f",};
		return HexCd[B>>4]+HexCd[B&0xf];

	}
	
	
	/** Print debugging information */
	public void printDebug() {
		System.out.print(
			"  PC= $"+HexStr(PC>>8)+HexStr(PC&0xff)+
			"  A = $"+HexStr(A)+
			"  X = $"+HexStr(X)+
			"  Y = $"+HexStr(Y)+" - Flags: "
		);
		if ((P&C_FLAG)!=0) System.out.print("C"); else System.out.print(" ");
		if ((P&Z_FLAG)!=0) System.out.print("Z"); else System.out.print(" ");
		if ((P&I_FLAG)!=0) System.out.print("I"); else System.out.print(" ");
		if ((P&D_FLAG)!=0) System.out.print("D"); else System.out.print(" ");
		if ((P&R_FLAG)!=0) System.out.print("R"); else System.out.print(" ");
		if ((P&V_FLAG)!=0) System.out.print("V"); else System.out.print(" ");
		if ((P&V_FLAG)!=0) System.out.print("N"); else System.out.print(" ");
		System.out.println("\t Instr: "+DAsm(PC));
	}


} // end of class J6502

