/*********************************************************************************
 
	The Machine Driver for the 6502 machines

	by 

	Norbert Kehrer

	Written in August 1998.

**********************************************************************************

*/


import java.awt.*;
import java.awt.image.*;
import java.awt.image.MemoryImageSource;
import java.awt.image.ColorModel;
import java.awt.event.*;
import java.util.*;
import java.io.*;
import java.net.*;
import java.applet.*;
import java.net.*;


// ***
// *** Class for 6502 machines
// ***

public class J6502Machine extends J6502 {

	// ROM files and sound files

	public URL    gameURL;		// Will be set by Applet
	public String GameDir = "";	// Set in subclass!


	// Video hardware

	public Applet		callingApplet = null;   // Will be set by the applet

	public Graphics 	parentGraphics = null;	// applet 
	public Graphics 	canvasGraphics = null;	// emulator window 
	public Graphics 	bufferGraphics = null;	// double buffering graphics - main buffer
	public Graphics 	buffer1Graphics = null;	// double buffering graphics - buffer 1
	public Graphics 	buffer2Graphics = null;	// double buffering graphics - buffer 2
	public Image		bufferImage    = null;  // double buffering image - main buffer
	public Image		buffer1Image    = null; // double buffering image - buffer 1
	public Image		buffer2Image    = null; // double buffering image - buffer 2
	
	public Applet		applet = null;          // the applet
	
	public Container	parent		= null; // the container
	public Canvas		canvas		= null; // the main screen


	// Input devices
	
	public J6502KeyListener	machineKeyListener;	// the listener for keyboard events
	Hashtable inPortHT;				// the hash table for input ports
	final static boolean PORT_ACTIVE_HIGH = true;	// Constant
	final static boolean PORT_ACTIVE_LOW =  false;	// Constant
	final static int DIGITAL_PORT =  1;		// Constant
	final static int ANALOG_PORT =   2;		// Constant


	// Frame control and skipping for speed-up

	public int  frameSkip = 1;
	public long frameCounter  = 0;


	// Profiling 

	public	long	profTimer = 0;
	public	long	loop6502Counter = 0;

	
 
	/***** Machine creation ***/

	public J6502Machine ( 	Container _parent, double clockFrequencyInMHz,
				int xScreen, int yScreen ) {

		// *** Set clock frequency

                super(clockFrequencyInMHz);
		
	
		// *** Create video hardware and input devices (keyboard)

		parent = _parent;
	
		parent.add( canvas = new Canvas() );
		canvas.setSize(new Dimension(xScreen, yScreen));

		canvas.setVisible(true);
		canvas.requestFocus();

		canvasGraphics = canvas.getGraphics();
		
		bufferImage = parent.createImage(xScreen , yScreen);	// main buffer
		bufferGraphics = bufferImage.getGraphics();

		buffer1Image = parent.createImage(xScreen , yScreen);	// buffer 1
		buffer1Graphics = buffer1Image.getGraphics();

		buffer2Image = parent.createImage(xScreen , yScreen);	// buffer 2
		buffer2Graphics = buffer2Image.getGraphics();

		canvasGraphics.setColor (Color.black);
		bufferGraphics.setColor (Color.black);
		buffer1Graphics.setColor (Color.black);
		buffer2Graphics.setColor (Color.black);
		canvasGraphics.fillRect(0, 0, xScreen, yScreen );
		bufferGraphics.fillRect(0, 0, xScreen, yScreen);
		buffer1Graphics.fillRect(0, 0, xScreen, yScreen);
		buffer2Graphics.fillRect(0, 0, xScreen, yScreen);
		canvasGraphics.setColor (Color.yellow);
		bufferGraphics.setColor (Color.yellow);
		buffer1Graphics.setColor (Color.yellow);
		buffer2Graphics.setColor (Color.yellow);


		// *** Create input devices (keyboard)

		inPortHT = new Hashtable();

		machineKeyListener = new J6502KeyListener(this);
		canvas.addKeyListener(machineKeyListener);


		// *** Initialize memory	
	
		for (int ii=0; ii<65536; ii++) mem[ii]=0xff;

	}



	
	/***** Load program ROMs ***/

	public void loadRomChip (int start, int rsize, String fileName) {
		URL baseURL = gameURL;
		String fileNam = GameDir + "/" +fileName;
		byte buffer[] = new byte [rsize];
		int readbytes=0;
		System.out.print("Reading ROM ... " + fileNam);
		try {
			InputStream is;
			if (baseURL != null) {
				URL romURL = new URL (baseURL, fileNam);
				is = romURL.openStream();
			}
			else {
				File romFile = new File (fileNam);
				is = new FileInputStream (romFile);
			}
			BufferedInputStream bis = new BufferedInputStream (is, rsize);
			int n=rsize;
			int toRead = rsize; 
			while ( toRead > 0 ) {
				int nRead = bis.read( buffer, n-toRead, toRead );
				toRead -= nRead;
				readbytes += nRead;
			}
			System.out.println (" ... "+readbytes+" bytes.");				
			for (int i=0;i<rsize;i++) {mem[i+start]=(buffer[i]+256)&0xff;}
			}
		catch (Exception e) {
			System.out.println ("Error loading ROM.");
		}
	}


	/***** Load sample files for sound hardware ***/

	public AudioClip loadSoundFile (String fileName) {

		URL baseURL = gameURL;
		String fileNam = GameDir + "/" + fileName;

		System.out.println("Loading sound file ... " + fileNam);

		AudioClip ac;

		ac = callingApplet.getAudioClip(baseURL, fileNam);
		ac.play();
		ac.stop();
		return ac;
	}


	/***** Load image files for graphically enchanced objects ***/

	public Image loadImageFile (String fileName) {

		URL baseURL = gameURL;
		String fileNam = GameDir + "/" + fileName;

		System.out.println("Loading image file ... " + fileNam);

		Image im = callingApplet.getImage(baseURL, fileNam);

		// Filter to make "black" the transparent color
		ImageFilter transpFilter = new TransparencyFilter();
		Image imF = callingApplet.createImage(
				new FilteredImageSource(im.getSource(), transpFilter));

		return imF;
	}


	/***** Input port bit definition ***/

	class J6502InPortBit extends Object {

		public int	portType;   // digital or analog
		public boolean	activeHigh;
		public int	memAddr;
		public int	setValue;
		public int	clrValue;

		public int	chgValue;		// Analog
		public int	analogSpeed;		// Analog
		public long	timeofLastChange;	// Analog

		public J6502InPortBit( int pt, boolean ah, int ma, int sv, int cv, int as ) {
			sv &= 0xff;
			portType = pt;
			activeHigh = ah;
			memAddr = ma;
			setValue = sv;
			clrValue = (~sv) & 0xff;
			chgValue = cv;
			analogSpeed = as;
			timeofLastChange = 0;
		}
	}


	/***** Define keyboard input - called by subclasses ***/

	public void defineKey (int keyCode, boolean activeHigh, int memAddr, int setValue) {

		J6502InPortBit pb = new J6502InPortBit (DIGITAL_PORT, activeHigh, memAddr, setValue, 0, 0);
		Integer keyCodeObj = new Integer(keyCode);
		inPortHT.put(keyCodeObj, pb);	
	
		if (activeHigh) mem[memAddr] = (~setValue) & 0xff;
		else		mem[memAddr] =   setValue  & 0xff;
		
	}

	/***** Define keyboard input for analog ports - called by subclasses ***/

	public void defineAnalogKey (int keyCode, int memAddr, int changeValue, int analogSpeed) {

		J6502InPortBit pb = new J6502InPortBit (ANALOG_PORT, PORT_ACTIVE_HIGH, memAddr, 0, changeValue, analogSpeed);
		Integer keyCodeObj = new Integer(keyCode);
		inPortHT.put(keyCodeObj, pb);	
	
		mem[memAddr] = 0x00;
		
	}


	/***** Process events from user interface ***/

	class J6502KeyListener implements KeyListener {
	
		J6502Machine theMachine;

		public J6502KeyListener( J6502Machine machine ) {
			theMachine = machine;
		}

		public void keyPressed( KeyEvent e ) {
			theMachine.handlePortKey( KeyEvent.KEY_PRESSED, e.getKeyCode() ); 
		}

		public void keyReleased( KeyEvent e ) {
			theMachine.handlePortKey( KeyEvent.KEY_RELEASED, e.getKeyCode() );	
		}

		public void keyTyped( KeyEvent e ) {
		}
	}



	/***** Filter to make "black" the transparent color in an image ***/

	class TransparencyFilter extends RGBImageFilter {
	
		public TransparencyFilter() {
			canFilterIndexColorModel = true;
		}

		public int filterRGB(int x, int y, int rgb) {
			if ( (rgb & 0x00ffffff) == 0 )
				return (0x00000000);	// transparent black
			else
				return (rgb);		// original color
		}
	
	}




	/***** Input port key handling ***/

	public void handlePortKey( int keyAction, int keyCode) {
		J6502InPortBit portBit;
		long timeOfThisChange;

		Integer keyCodeObj = new Integer(keyCode);

		portBit = (J6502InPortBit) inPortHT.get(keyCodeObj);
		if (portBit != null) {
			if (keyAction == KeyEvent.KEY_PRESSED) {
				switch (portBit.portType) {
					case DIGITAL_PORT:
						if (portBit.activeHigh) mem[portBit.memAddr] |= portBit.setValue;
						else			mem[portBit.memAddr] &= portBit.clrValue;
						break;
					case ANALOG_PORT:
//						timeOfThisChange = System.currentTimeMillis();
//						if ( (timeOfThisChange - portBit.timeOfLastChange) > portBit.analogSpeed ) {
//							portBit.timeOfLastChange = timeOfThisChange;
							mem[portBit.memAddr] += portBit.chgValue;
							if (mem[portBit.memAddr] > 0xff) mem[portBit.memAddr] = 0xff;
							if (mem[portBit.memAddr] < 0x00) mem[portBit.memAddr] = 0x00;
//						}
						break;
				}
			}
			if (keyAction == KeyEvent.KEY_RELEASED) {
				if (portBit.activeHigh) mem[portBit.memAddr] &= portBit.clrValue;
				else			mem[portBit.memAddr] |= portBit.setValue;
			}
		}

		// Keys for frame skip control
		if (keyAction == KeyEvent.KEY_PRESSED) {
			if ( keyCode == KeyEvent.VK_SUBTRACT ) {
					frameSkip -= 1; // Decrease frame skip
					if (frameSkip < 1) frameSkip = 1;
					System.out.println ("Frame skip: "+frameSkip);
			}
			if ( keyCode == KeyEvent.VK_ADD ) {
					frameSkip += 1; // Increase frame skip
					System.out.println ("Frame skip: "+frameSkip);
			}
		}


		// Call for driver specific actions on keyboard events
		handleKey( keyAction, keyCode);	// further keys in subclasses
	}


	/***** Further keyboard handling - This method may be defined in subclasses ***/

	public void handleKey( int keyAction, int keyCode) {
	}


	/***** Load ROMs - This method must be defined in subclasses ***/

	public void loadRom () {
	}
		

	/***** Load sounds - This method may be defined in subclasses ***/

	public void loadSound () {
	}


	/***** Load images - This method may be defined in subclasses ***/

	public void loadImages () {
	}


	/***** Debugging and profiling ***/

	public void timerOn () {
		profTimer = System.currentTimeMillis();
	}


	public void timerOff (String info) {
		profTimer = System.currentTimeMillis() - profTimer;
		System.out.println(info + profTimer);
	}

		
} // end of class J6502Machine



// ***
// *** Applet for 6502 machines
// ***


class J6502MachineApplet extends Applet implements Runnable {

	J6502Machine	j6502machine = null;
	Thread		thread	= null;
	URL 		baseURL = null;
	
	
	public void init() {
		try {
			baseURL = getDocumentBase();
		}
		catch (Exception e) {
			baseURL = null;
		}	
	}
	
	public void start() {
		if ( thread == null ) {
			thread = new Thread( this, "6502 Machine" );
			thread.start();
		}
	}
	
	
	public void stop() {
		if ( thread != null ) {
			thread.stop();
			thread = null;
		}
	}
	
	
	/** Read applet parameters and start the emulation. */
	public void run() {
		if ( j6502machine == null )
			j6502machine = createMachine();

		if ( j6502machine != null ) {
			j6502machine.Debugging = false;
			j6502machine.gameURL = baseURL;
			j6502machine.callingApplet = this;
			j6502machine.loadRom();	
			j6502machine.loadSound();	
			j6502machine.loadImages();	
			j6502machine.reset();
			j6502machine.execute();
		}
	}


	// Define in subclasses!
	public J6502Machine createMachine () {
		return null;
	}
}





