/*********************************************************************************
 
	The Asteroids Deluxe Emulator in Java

	by 

	Norbert Kehrer

	August 1998

**********************************************************************************

	Information sources:
		- Asteroids emulator of the MAME project - Thanks for the sources
		- Asteroids emulator by Chris Pile       - Thanks for the documentation

	
	Asteroids Deluxe is an Atari vector game from 1980

*/


import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;
import java.net.*;
import java.applet.*;


// ***
// *** ATARI Asteroids Deluxe Hardware
// ***

class AsteroidsDeluxe extends J6502Machine {

	// Video hardware
	
	public AtariDVG dvg;		// Connected Atari vector processor

	static int xScreen = 512;	// screen size
	static int yScreen = 512;


	// Memory bank switch

	static int asteroid_bank = 0;
	static int asteroid_newbank;

	
	// Timing for correct speed

	static	long timePerFrame  = 20; // Speed limit is 50 frames per second
	static	long thisFrameDuration = 0;
	static	long lastFrameTime = 0;



	// Sounds

	static AudioClip beat1SFX       = null;
	static AudioClip beat2SFX       = null;
	static AudioClip largeExplSFX   = null;
	static AudioClip mediumExplSFX  = null;	
	static AudioClip smallExplSFX   = null;	
	static AudioClip largeSaucerSFX = null;
	static AudioClip smallSaucerSFX = null;
	static AudioClip thrustSFX      = null;
	static AudioClip ufoFireSFX     = null;
	static AudioClip yourFireSFX    = null;
	static AudioClip extraShipSFX   = null;

	static int lastPlayerFire, thisPlayerFire, lastUfoFire, thisUfoFire;	


 
	/***** Asteroids Deluxe creation ***/

	public AsteroidsDeluxe ( Container _parent ) {

		// *** Asteroids Deluxe runs at 1.50 Mhz  - but we use 0.3125 MHz for more exact NMIs

                super( _parent, 0.3125, xScreen, yScreen );
		

		// *** Set game directory

		GameDir = "astdelux";


		// *** Initialize memory	

		mem[0x2002] = 0; // Vector halt
		mem[0x2007] = 0; // Self test


		// *** Pre-set DIP switch memory locations
	
		mem[0x2800] = 0xfc | 0x02; // 2 = 1 Coin 1 Play,  0 = free play
		mem[0x2801] = 0xfc | 0x00; // Unknown
		mem[0x2802] = 0xfc | 0x01; // 4-ship game
		mem[0x2803] = 0xfc | 0x00; // English


		// *** Define keys of input ports

		defineKey (KeyEvent.VK_LEFT,	PORT_ACTIVE_HIGH, 0x2407, 0x80); // Rotate left
		defineKey (KeyEvent.VK_RIGHT,	PORT_ACTIVE_HIGH, 0x2406, 0x80); // Rotate right
		defineKey (KeyEvent.VK_CONTROL,	PORT_ACTIVE_HIGH, 0x2004, 0x80); // Fire
		defineKey (KeyEvent.VK_UP,	PORT_ACTIVE_HIGH, 0x2405, 0x80); // Thrust
		defineKey (KeyEvent.VK_SPACE,	PORT_ACTIVE_HIGH, 0x2003, 0x80); // Shield
		defineKey (KeyEvent.VK_3,	PORT_ACTIVE_HIGH, 0x2400, 0x80);
		defineKey (KeyEvent.VK_4,	PORT_ACTIVE_HIGH, 0x2401, 0x80);
		defineKey (KeyEvent.VK_5,	PORT_ACTIVE_HIGH, 0x2402, 0x80);
		defineKey (KeyEvent.VK_1,	PORT_ACTIVE_HIGH, 0x2403, 0x80);
		defineKey (KeyEvent.VK_2,	PORT_ACTIVE_HIGH, 0x2404, 0x80);


		// *** Clear input port memory locations

		for (int i = 0; i < 8; i++) {	
			mem[0x2000+i] = 0x7f;
			mem[0x2400+i] = 0x7f;
		}


		// *** Clear EAROM (Hi scores)

		for (int i = 0; i < 0x40; i++) mem[0x2c40+i] = 0x00;


		// *** Create the DVG vector engine

		dvg = new AtariDVG(mem, 0x4000 /* vector RAM */, bufferGraphics, xScreen, yScreen);

	}



	/***** Memory access ***/

	public final void poke( int addr, int newByte ) {


		switch ( addr ) {

		// *** Start the DVG engine and draw the screen

		case 0x3000:
			frameCounter++;
			if ((frameCounter % frameSkip) == 0) {
				dvg.dvgRun();
				canvasGraphics.drawImage(bufferImage,0,0,null);
			}

			// Wait when too fast
			thisFrameDuration = System.currentTimeMillis() - lastFrameTime;
			if (thisFrameDuration < timePerFrame) {
				try { Thread.sleep(timePerFrame - thisFrameDuration); } 
				catch ( Exception ignored ) {}
			}
			lastFrameTime = System.currentTimeMillis();
			break;


		// *** Memory bank switch

		case 0x3c04:
			asteroid_newbank = (newByte >> 7) & 0x01;
			if (asteroid_bank != asteroid_newbank) {
				/* Perform bankswitching on page 2 and page 3 */
				int temp;
				int i;
				asteroid_bank = asteroid_newbank;
				for (i = 0; i < 0x100; i++) {
					temp = mem[0x200 + i];
					mem[0x200 + i] = mem[0x300 + i];
					mem[0x300 + i] = temp;
				}
			}
			break;


		// *** Sound circuits control - Pokey sounds are not emulated

		case 0x2405: // Thrust
			if ((newByte&0x80) != 0) thrustSFX.loop();	
			else			 thrustSFX.stop();	
 			break;	


		case 0x3600: // Explosions
			if (newByte == 0x3d) largeExplSFX.play(); 
			if (newByte == 0xfd) mediumExplSFX.play(); 
			if (newByte == 0xbd) smallExplSFX.play(); 
			break;


		} // end case


		// *** Normal RAM

		if ( addr < 0x6000 )         // 0x6000 - 0xffff ROM chips
			mem [addr]=newByte;	
		
		return;
		
	}
	

	public final int peek( int addr ) {

		// *** Pokey Random Number Generator

		if ( addr == 0x2c0a ) return (int) (Math.random()*255);


		// *** Rest of memory

		return mem[addr];
	}
	
	
	/***** Load program ROMs ***/

	public void loadRom () {
		loadRomChip (0x6000,0x0800,"036430.02");  // code
		loadRomChip (0x6800,0x0800,"036431.02");  // code
		loadRomChip (0x7000,0x0800,"036432.02");  // code
		loadRomChip (0x7800,0x0800,"036433.03");  // code
		loadRomChip (0xf800,0x0800,"036433.03");  // reload code
		loadRomChip (0x4800,0x0800,"036800.02");  // vector ROM
		loadRomChip (0x5000,0x0800,"036799.01");  // vector ROM
	};


	/***** Load files for sound hardware ***/

	public void loadSound () {

		beat1SFX =     	loadSoundFile ("Beat1.au");		
		beat2SFX =     	loadSoundFile ("Beat2.au");		
		largeExplSFX = 	loadSoundFile ("Largexpl.au");		
		mediumExplSFX =	loadSoundFile ("Medumxpl.au");		
		smallExplSFX =	loadSoundFile ("Smallxpl.au");		
		largeSaucerSFX= loadSoundFile ("Lsaucer.au");		
		smallSaucerSFX=	loadSoundFile ("Ssaucer.au");		
		thrustSFX = 	loadSoundFile ("Thrust.au");		
		ufoFireSFX = 	loadSoundFile ("Ufofire.au");		
		yourFireSFX = 	loadSoundFile ("Yourfire.au");		
		extraShipSFX = 	loadSoundFile ("Extraman.au");		
	}



		
	/***** Periodic handler to drive the 6502 interrupts ***/

	public final int Loop6502() {
 	
		loop6502Counter++;
	
		/*
		 * We catch the following busy loop:
		 * 6014 bit $2002
		 * 6017 bmi $6014
		 */

		if ((PC >= 0x6014) && (PC <= 0x6018)) PC = 0x6819;


		// We update the 2 KHz Source

		if ((mem[0x2001]&0x80) != 0x00) 	mem[0x2001]&=0x7f;
		else 					mem[0x2001]|=0x80;


		// We trigger an interrupt

	 	return super.INT_NMI;
	}

		
		
} // end of class AsteroidsDeluxe



// ***
// *** ATARI Asteroids Deluxe Applet
// ***


public class JAsteroidsDeluxe extends J6502MachineApplet implements Runnable {
	
	// *** Create the Asteroids Deluxe machine
	public J6502Machine createMachine ( ) {

		return ( new AsteroidsDeluxe(this) ) ;

	}	
}





