/* 
	The Astro Fighter Emulator in Java

	by 

	Norbert Kehrer


	Lots of info from M.A.M.E.
*/


/** Hardware spec for Astro Fighter

Memory map:

0x0000 - 0x01ff: RAM  
0x4000 - 0x5fff: Video RAM  
0x8003 - 0x8003: Color RAM  
0xa000 - 0xa000: Input port 0 
0xa001 - 0xa001: Input port 1 
0xd000 - 0xffff: Program ROM 

*/


import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;
import java.net.*;
import java.applet.*;


class Astrof extends J6502Machine {


	static int xScreen = 512;	// screen size
	static int yScreen = 512;


	/** DIP switches, VBLANK and NMI*/

	private int DIP_SWITCH 	= 0x00; // Standard settings
	private int IP_VBLANK	= 0x80; // VBLANK bit

	private int VBLANK 		= 0x80;
	private boolean isVBLANK 	= false;

	private boolean DO_NMI 	= false;

	// Timing for correct speed

	static	long timePerFrame  = 20; // Speed limit is 50 frames per second
	static	long thisFrameDuration = 0;
	static	long lastFrameTime = 0;


	// intermediary data
	private int i,x,y,offset,data;

	// Colors: Alpha - Red - Green - Blue
	Color BLACK		= new Color (0xff000000);
	Color GREY		= new Color (0xffa0a0a0);
	Color GREEN		= new Color (0xff00ff00);
	Color PURPLE		= new Color (0xffff00ff);
	Color CYAN		= new Color (0xff00ffff);
	Color WHITE		= new Color (0xffffffff);
	Color RED		= new Color (0xffff0000);
        Color YELLOW            = new Color (0xffffff00);
	

	Color drawColor = BLACK;
	Color zeroColor = BLACK;

	Color ColorTable[] = {
		BLACK, GREY, GREEN, GREEN, PURPLE, CYAN, WHITE, RED, YELLOW
	};



	/** Astro Fighter creation */
	public Astrof ( Container _parent ) {

		// Astro Fighter runs at 0.66 Mhz  - ?? 
                super(_parent, 0.66, xScreen, yScreen );

		
		// *** Set game directory

		GameDir = "astrof";


		// *** Define keys of input ports

		defineKey (KeyEvent.VK_LEFT,	PORT_ACTIVE_LOW, 0xa000, 0x48);
		defineKey (KeyEvent.VK_RIGHT,	PORT_ACTIVE_LOW, 0xa000, 0x24);
		defineKey (KeyEvent.VK_CONTROL,	PORT_ACTIVE_LOW, 0xa000, 0x90);
		defineKey (KeyEvent.VK_1,	PORT_ACTIVE_LOW, 0xa000, 0x01);
		defineKey (KeyEvent.VK_2,	PORT_ACTIVE_LOW, 0xa000, 0x02);


		// *** Initialize memory		
		for (int ii=0; ii<65536; ii++) mem[ii]=0xff;
		
	}


	/** More keyboard handling - Coin insertion triggers NMI ***/

	public void handleKey( int keyAction, int keyCode) {
		if ( (keyAction == KeyEvent.KEY_PRESSED) && (keyCode == KeyEvent.VK_3 ) )
			DO_NMI = true;
	}



	/** Memory access */

	public void poke( int addr, int newByte ) {

                // *** Video RAM:

                if (addr >=  0x4000 && addr <= 0x5fff && mem[addr] != newByte) {

			offset = addr - 0x4000;
			data = newByte;

			x = offset & 0xff; 
			y = (1 + (offset >> 8))<<3;
			for (i = y; i > y-8 ; i--) {
				if (i < 256) {
					if ((data & 0x80) != 0)	plotPixel(x,i,drawColor);
					else 			plotPixel(x,i,zeroColor);
				}
				data <<= 1;
			}
		}


                // *** Color RAM:

                if (addr ==  0x8003) {
			drawColor = ColorTable[(newByte >> 1) + 1];
		}


                // *** Normal RAM:

		if ( addr < 0xD000 ) { 	// 0xD000 - 0xffff Program ROM 
			mem [addr]=newByte;	
		}
		
		return;
		
	}
	
	
	public int peek( int addr ) {

		// *** Dip switches and VBLANK bit 

		if(addr == 0xa001) {
			i = (VBLANK | DIP_SWITCH);
			VBLANK = 0;
			return i;
		}


		// *** Normal memory

		return mem[addr];
				
	}
	
	
	/****** Load program ROMs */
	public void loadRom () {
		loadRomChip (0xd000,0x0400,"kei2");	
		loadRomChip (0xd400,0x0400,"keii");	
		loadRomChip (0xd800,0x0400,"kei0");	
		loadRomChip (0xdc00,0x0400,"ke9");	
		loadRomChip (0xe000,0x0400,"ke8");	
		loadRomChip (0xe400,0x0400,"ke7");	
		loadRomChip (0xe800,0x0400,"ke6");	
		loadRomChip (0xec00,0x0400,"ke5");	
		loadRomChip (0xf000,0x0400,"ke4");	
		loadRomChip (0xf400,0x0400,"ke3");	
		loadRomChip (0xf800,0x0400,"ke2");	
		loadRomChip (0xfc00,0x0400,"kei");	

	};



		

		
	/** Periodic handler to drive the 6502 interrupts*/

	public final int Loop6502() {

		loop6502Counter++;
		
		if (loop6502Counter % frameSkip == 0)  screenRefresh ();

		if (DO_NMI) {
			DO_NMI = false;
			return super.INT_NMI;
		}
		else { 	return super.INT_NONE; }
	}




	/** Screen drawing */	
	
	public void screenRefresh () {

		// Refresh VBLANK bit 

		VBLANK = IP_VBLANK;
		canvasGraphics.drawImage(bufferImage,0,0,null);

		// Wait when too fast
		thisFrameDuration = System.currentTimeMillis() - lastFrameTime;
		if (thisFrameDuration < timePerFrame) {
			try { Thread.sleep(timePerFrame - thisFrameDuration); } 
			catch ( Exception ignored ) {}
		}
		lastFrameTime = System.currentTimeMillis();


	}


	/** Pixel plotter */	
	
       public void plotPixel (int x, int y, Color pixelColor) {
		bufferGraphics.setColor(pixelColor);
		bufferGraphics.fillRect(x<<1,y<<1,2,1);
						
	}

		
} // end of class Astrof


// ***
// *** Astro Fighter Applet
// ***


public class JAstrof extends J6502MachineApplet implements Runnable {
	
	// *** Create the Astro Fighter machine
	public J6502Machine createMachine ( ) {

		return ( new Astrof(this) ) ;

	}	
}



