/* 
        The Centipede Emulator in Java

	by 

	Norbert Kehrer

	Written in July 1999.

	Centipede was a famous Atari arcade game from 1980.

        All the info from the MAME project.
	Thank you for your great work.

*/


/** Hardware spec for Centipede - taken from the MAME source


              Centipede Memory map and Dip Switches
              -------------------------------------

Memory map for Centipede directly from the Atari schematics (1981).

 Address  R/W  D7 D6 D5 D4 D3 D2 D1 D0   Function
--------------------------------------------------------------------------------------
0000-03FF       D  D  D  D  D  D  D  D   RAM
--------------------------------------------------------------------------------------
0400-07BF       D  D  D  D  D  D  D  D   Playfield RAM
07C0-07CF       D  D  D  D  D  D  D  D   Motion Object Picture
07D0-07DF       D  D  D  D  D  D  D  D   Motion Object Vert.
07E0-07EF       D  D  D  D  D  D  D  D   Motion Object Horiz.
07F0-07FF             D  D  D  D  D  D   Motion Object Color
--------------------------------------------------------------------------------------
0800       R    D  D  D  D  D  D  D  D   Option Switch 1 (0 = On)
0801       R    D  D  D  D  D  D  D  D   Option Switch 2 (0 = On)
--------------------------------------------------------------------------------------
0C00       R    D           D  D  D  D   Horizontal Mini-Track Ball tm Inputs
           R       D                     VBLANK  (1 = VBlank)
           R          D                  Self-Test  (0 = On)
           R             D               Cocktail Cabinet  (1 = Cocktail)
0C01       R    D  D  D                  R,C,L Coin Switches (0 = On)
           R             D               SLAM  (0 = On)
           R                D            Player 2 Fire Switch (0 = On)
           R                   D         Player 1 Fire Switch (0 = On)
           R                      D      Player 2 Start Switch (0 = On)
           R                         D   Player 1 Start Switch (0 = On)

0C02       R    D           D  D  D  D   Vertical Mini-Track Ball tm Inputs
0C03       R    D  D  D  D               Player 1 Joystick (R,L,Down,Up)
           R                D  D  D  D   Player 2 Joystick   (0 = On)
--------------------------------------------------------------------------------------
1000-100F R/W   D  D  D  D  D  D  D  D   Custom Audio Chip
1404       W                D  D  D  D   Playfield Color RAM
140C       W                D  D  D  D   Motion Object Color RAM
--------------------------------------------------------------------------------------
1600       W    D  D  D  D  D  D  D  D   EA ROM Address & Data Latch
1680       W                D  D  D  D   EA ROM Control Latch
1700       R    D  D  D  D  D  D  D  D   EA ROM Read Data
--------------------------------------------------------------------------------------
1800       W                             IRQ Acknowledge
--------------------------------------------------------------------------------------
1C00       W    D                        Left Coin Counter (1 = On)
1C01       W    D                        Center Coin Counter (1 = On)
1C02       W    D                        Right Coin Counter (1 = On)
1C03       W    D                        Player 1 Start LED (0 = On)
1C04       W    D                        Player 2 Start LED (0 = On)
1C07       W    D                        Track Ball Flip Control (0 = Player 1)
--------------------------------------------------------------------------------------
2000       W                             WATCHDOG
2400       W                             Clear Mini-Track Ball Counters
--------------------------------------------------------------------------------------
2000-3FFF  R                             Program ROM
--------------------------------------------------------------------------------------

-EA ROM is an Erasable Reprogrammable rom to save the top 3 high scores
  and other stuff.


 Dip switches at N9 on the PCB

 8    7    6    5    4    3    2    1    Option
-------------------------------------------------------------------------------------
                              On   On    English $
                              On   Off   German
                              Off  On    French
                              Off  Off   Spanish
-------------------------------------------------------------------------------------
                    On   On              2 lives per game
                    On   Off             3 lives per game $
                    Off  On              4 lives per game
                    Off  Off             5 lives per game
-------------------------------------------------------------------------------------
                                         Bonus life granted at every:
          On   On                        10,000 points
          On   Off                       12.000 points $
          Off  On                        15,000 points
          Off  Off                       20,000 points
-------------------------------------------------------------------------------------
     On                                  Hard game difficulty
     Off                                 Easy game difficulty $
-------------------------------------------------------------------------------------
On                                       1-credit minimum $
Off                                      2-credit minimum
-------------------------------------------------------------------------------------

$ = Manufacturer's suggested settings


 Dip switches at N8 on the PCB

 8    7    6    5    4    3    2    1    Option
-------------------------------------------------------------------------------------
                              On   On    Free play
                              On   Off   1 coin for 2 credits
                              Off  On    1 coin for 1 credit $
                              Off  Off   2 coins for 1 credit
-------------------------------------------------------------------------------------
                    On   On              Right coin mech X 1 $
                    On   Off             Right coin mech X 4
                    Off  On              Right coin mech X 5
                    Off  Off             Right coin mech X 6
-------------------------------------------------------------------------------------
               On                        Left coin mech X 1 $
               Off                       Left coin mech X 2
-------------------------------------------------------------------------------------
On   On   On                             No bonus coins $
On   On   Off                            For every 2 coins inserted, game logic
                                          adds 1 more coin
On   Off  On                             For every 4 coins inserted, game logic
                                          adds 1 more coin
On   Off  Off                            For every 4 coins inserted, game logic
                                          adds 2 more coin
Off  On   On                             For every 5 coins inserted, game logic
                                          adds 1 more coin
Off  On   Off                            For every 3 coins inserted, game logic
                                          adds 1 more coin
-------------------------------------------------------------------------------------
$ = Manufacturer's suggested settings

*/



import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;
import java.net.*;
import java.applet.*;


class Centiped extends J6502Machine {

	/** Creator applet */

	J6502MachineApplet myCreator = null;


	/** Screen size */

	static int xScreen = 512;	// screen size
	static int yScreen = 512;

	// Timing for correct speed

	static	long timePerFrame  = 20; // Speed limit is 50 frames per second
	static	long thisFrameDuration = 0;
	static	long lastFrameTime = 0;


	/** Character set */

	int[] charPixels;
	MemoryImageSource charSource;
	IndexColorModel charCM;
	byte rChar[] = {(byte)0x00, (byte)0xff, (byte)0x00, (byte)0x00}; // red components - default
	byte gChar[] = {(byte)0x00, (byte)0x00, (byte)0xff, (byte)0x00}; // red components - default
	byte bChar[] = {(byte)0x00, (byte)0x00, (byte)0x00, (byte)0xff}; // red components - default

	Image charSet; 


	/** Motion objects */

	IndexColorModel mobCM;
	byte rMob[] = {(byte)0x00, (byte)0xff, (byte)0x00, (byte)0xff}; // red components - default
	byte gMob[] = {(byte)0x00, (byte)0xff, (byte)0xff, (byte)0x00}; // red components - default
	byte bMob[] = {(byte)0x00, (byte)0x00, (byte)0x00, (byte)0xff}; // red components - default
	MemoryImageSource mobSource;

	Image mobSet; 


	/** Pallette RAM dirty */

	boolean newColors = true;


	/** Dirty buffer for Video RAM */

	int dirtyCounter = 0;
	int dirtyBuffer[] = new int[0x400*4];	// Buffer is 4 times the video ram to be
						// rather sure (CPU can 4 times change the

	/** VBLANK */

	private int VBLANK 		= 0xff;
	private boolean isVBLANK 	= false;

	private int PORT_0_VBLNK = (0x20|VBLANK|0x81);
	private int PORT_0_NOVBL = (0x20|0x81);


	/** Centiped creation */

	public Centiped ( J6502MachineApplet _parent ) {

		// *** Centipede runs at 1 MHz 
                super(_parent, 1.00, xScreen, yScreen );
		

		// *** Remember creator applet
		myCreator = _parent;
		

		// *** Set game directory
		GameDir = "centiped";


		// *** Define keys of input ports
		defineKey (KeyEvent.VK_UP,	PORT_ACTIVE_LOW, 0x0c03, 0x11); // Up
		defineKey (KeyEvent.VK_DOWN,	PORT_ACTIVE_LOW, 0x0c03, 0x22); // Down
		defineKey (KeyEvent.VK_LEFT,	PORT_ACTIVE_LOW, 0x0c03, 0x44);	// Left
		defineKey (KeyEvent.VK_RIGHT,	PORT_ACTIVE_LOW, 0x0c03, 0x88); // Right
		defineKey (KeyEvent.VK_CONTROL,	PORT_ACTIVE_LOW, 0x0c01, 0x0c); // Fire button 
		defineKey (KeyEvent.VK_1,	PORT_ACTIVE_LOW, 0x0c01, 0x01); // Start 1
		defineKey (KeyEvent.VK_2,	PORT_ACTIVE_LOW, 0x0c01, 0x02); // Start 2
		defineKey (KeyEvent.VK_3,	PORT_ACTIVE_LOW, 0x0c01, 0x20); // Coin 1
		defineKey (KeyEvent.VK_3,	PORT_ACTIVE_LOW, 0x0c01, 0x40); // Coin 3
		defineKey (KeyEvent.VK_4,	PORT_ACTIVE_LOW, 0x0c01, 0x80); // Coin 2


		// *** Initialize memory		
		for (int ii=0; ii<65536; ii++) mem[ii]=0xff;

		// *** Set dip switches
		mem[0x0800] = 0x54; // Atari suggestion
		mem[0x0801] = 0x02; // Atari suggestion


		// *** Initialize ports	
		mem[0x0c01] = 0xff;
		mem[0x0c03] = 0xff;


		// *** Reset video RAM buffer
		dirtyCounter = 0;

	}





	/** Memory access */

	public void poke( int addr, int newByte ) {

		// *** Video RAM:
		if( (addr>=0x0400) & (addr<=0x07bf) ) {
			dirtyBuffer[dirtyCounter++] = addr-0x400; // remember dirty position			
			mem[addr]=newByte;
			return;
		}


                // *** Pallette RAM  - characters:
		if (addr >=  0x1404 && addr <= 0x1407) newColors = true;

                // *** Pallette RAM  - motion objects:
		if (addr >=  0x140c && addr <= 0x140f) newColors = true;



                // *** Normal RAM:
		if ( addr < 0x2000 )  	// above this is Program ROM or nothing
			mem [addr]=newByte;	
		
		return;
		
	}
	
	public int peek( int addr ) {

		// *** VBLANK bit
		if (addr == 0x0c00) {
			// preliminary:
			if ( isVBLANK ) {
				isVBLANK = false;
				return PORT_0_VBLNK;
			}
			else {
				isVBLANK = true;
				return PORT_0_NOVBL;
			}
			
		}


 		// *** Pokey - Register 0x0A is the random number generator:
		if (addr == 0x100A) return (int) (Math.random()*255);


		// *** Normal memory:
		return mem[addr];
				
	}
	
	
	/****** Load cartridge file */
	public void loadRom () {

		// *** Load char prom into memory starting from 0x0000		
		loadRomChip (0x0000,0x0800,"centiped.211");	
		loadRomChip (0x0800,0x0800,"centiped.212");	

		// *** Load program proms into memory starting from 0x2000		
		loadRomChip (0x2000,0x0800,"centiped.307");	
		loadRomChip (0x2800,0x0800,"centiped.308");	
		loadRomChip (0x3000,0x0800,"centiped.309");	
		loadRomChip (0x3800,0x0800,"centiped.310");	
		loadRomChip (0xf800,0x0800,"centiped.310"); //Remapped for Reset, IRQ, and NMI vectors


		// *** Create character set
		int pp, bit1, bit2;

		charPixels = new int[256*8*8*2*2]; 

		for (int c = 0; c < 256; c++) {
			for (int i = 0; i < 8; i++) {
				for (int j = 0; j < 8; j++) {
					pp = j*256*8*2*2 + c*8*2 + i*2;
					bit1 = ((mem[      c*8+i]>>j)&0x01)<<1;
					bit2 =  (mem[(c+256)*8+i]>>j)&0x01;
					charPixels[pp]   = (bit1|bit2);
					charPixels[pp+1] = charPixels[pp];
				};
			};
		};

		charCM = new IndexColorModel(2, 4, rChar, gChar, bChar);

		charSource = new MemoryImageSource(256*8*2, 8*2, charCM, charPixels, 0, 256*8*2);
		charSource.setAnimated(true);
		charSource.setFullBufferUpdates(true);
		charSet = parent.createImage(charSource);


		// *** Create character set for motion objects
		// *** The images are the same as the character set images

		mobCM = new IndexColorModel(2, 4, rMob, gMob, bMob, 0);  // color 0 is transparent

		mobSource = new MemoryImageSource(256*8*2, 8*2, mobCM, charPixels, 0, 256*8*2);
		mobSource.setAnimated(true);
		mobSource.setFullBufferUpdates(true);
		mobSet = parent.createImage(mobSource);


		// *** Set dip switches
		mem[0x0800] = 0x54; // Atari suggestion
		mem[0x0801] = 0x02; // Atari suggestion


		// *** Initialize ports	
		mem[0x0c01] = 0xff;
		mem[0x0c03] = 0xff;


	};



		

		
	/** Periodic handler to drive the 6502 interrupts*/

	public final int Loop6502() {

		loop6502Counter++;
		
		if (loop6502Counter % frameSkip == 0)  screenRefresh ();

		/* VBLANK port*/
		isVBLANK = true;


		// ** Update colors
		if (newColors) {
			updateColors();
			newColors = false;
		}


		// ** Hardware and the Atari software ??

		if ((PC >= 0x3C0A) && (PC <= 0x3C15)) PC = 0x3C16;
		if ((PC >= 0x38A8) && (PC <= 0x38AF)) PC = 0x38B0;

		// ** IRQ 
		return super.INT_IRQ;

	}



	/** Screen drawing */	
	
	public void screenRefresh () {

		int i, j, pos, sx, sy, charcode, bx, by, ex, ey, x, y, spr, flipx;


		// *** Draw the playfield in buffer 1

		for(i=0; i<dirtyCounter; i++) { 

			pos = dirtyBuffer[i];

			charcode =  (mem[0x0400+pos] & 0x3F) + 0x40;

                        sx = (pos >> 5)<<4;		// i.e.  ((int)(pos / 32))<<4;
                        sy = (31-(pos & 0x1f))<<4;	// i.e. (31-(pos % 32))<<4;

			buffer1Graphics.drawImage(charSet,sx,sy,sx+16,sy+16,
					  	  (charcode<<4), 0,
						  (charcode<<4)+16, 16,
						  null);
		}

		dirtyCounter = 0;					// reset buffer

		bufferGraphics.drawImage(buffer1Image,0,0,null);	// draw buffer 1


		// *** Draw the motion objects in main buffer

		for (i = 0; i < 0x10; i++) {
			spr = mem[0x07c0 + i] & 0x3f;
			if ((spr & 0x01) != 0)	spr = (spr >> 1) + 64;
			else 			spr = (spr >> 1);

			flipx = (mem[ 0x07c0 + i]>>7);

                        sx = 480 - (mem[0x07d0 + i]<<1);
                        sy = 496 - (mem[0x07e0 + i]<<1);

			if (mem[0x07e0 + i] < 241) { // hide the top 2 lines
				if (flipx == 0)
					bufferGraphics.drawImage(mobSet,sx,sy,sx+32,sy+16,
							  	 (spr<<5), 0,
								 (spr<<5)+32, 16,
								  null);
				else
					bufferGraphics.drawImage(mobSet,sx+32,sy,sx,sy+16,
							  	 (spr<<5), 0,
								 (spr<<5)+32, 16,
								  null);
			}
		} 

		canvasGraphics.drawImage(bufferImage,0,0,null);		// draw frame

		// Wait when too fast
		thisFrameDuration = System.currentTimeMillis() - lastFrameTime;
		if (thisFrameDuration < timePerFrame) {
			try { Thread.sleep(timePerFrame - thisFrameDuration); } 
			catch ( Exception ignored ) {}
		}
		lastFrameTime = System.currentTimeMillis();

	}




	/** Generate color models from pallette RAM */

	public void updateColors () {
		int r, g, b, n, x, data;


		// *** Build pallettes

		for (n = 1; n<4; n++) {
			x=n;
			// ** Colors for characters
			data = mem[ 0x1404 + n ];
			if ((~data & 0x08) == 0) { /* luminance = 0 */
				r = 0xc0 * ((~data >> 0) & 1);
				g = 0xc0 * ((~data >> 1) & 1);
				b = 0xc0 * ((~data >> 2) & 1);
			}
			else { /* luminance = 1 */
				r = 0xff * ((~data >> 0) & 1);
				g = 0xff * ((~data >> 1) & 1);
				b = 0xff * ((~data >> 2) & 1);
			}
			if (n==1) x=2;
			if (n==2) x=1;
			rChar[x] = (byte)r;
			gChar[x] = (byte)g;
			bChar[x] = (byte)b;

		
			// ** Colors for sprites
			data = mem[ 0x140c + n ];
			if ((~data & 0x08) == 0) { /* luminance = 0 */
				r = 0xc0 * ((~data >> 0) & 1);
				g = 0xc0 * ((~data >> 1) & 1);
				b = 0xc0 * ((~data >> 2) & 1);
			}
			else { /* luminance = 1 */
				r = 0xff * ((~data >> 0) & 1);
				g = 0xff * ((~data >> 1) & 1);
				b = 0xff * ((~data >> 2) & 1);
			}
			if (n==1) x=2;
			if (n==2) x=1;
			rMob[x] = (byte)r;
			gMob[x] = (byte)g;
			bMob[x] = (byte)b;
		}		


		// *** Reconstruct color models

		charCM = new IndexColorModel(2, 4, rChar, gChar, bChar);
		charSource.newPixels(charPixels, charCM, 0, 256*8*2);

		mobCM = new IndexColorModel(2, 4, rMob, gMob, bMob, 0);
		mobSource.newPixels(charPixels, mobCM, 0, 256*8*2);


		// *** Complete screen is dirty now
                for (n = 0x0000; n < 0x03c0; n++) dirtyBuffer[n] = n;                    
		dirtyCounter = 0x3c0;

	}


} // end of class Centiped


// ***
// *** Centipede Applet
// ***


public class JCentiped extends J6502MachineApplet implements Runnable {
	
	// *** Create the Super breakout machine
	public J6502Machine createMachine ( ) {

		return ( new Centiped(this) ) ;

	}	
}



