/* 
        The Dominos Emulator in Java

	by 

	Norbert Kehrer

	Written in May and June 1999.

	Dominos was an Atari arcade game from 1978.

        All the info from the MAME project (Mike Balfour's source code).
	Thank you for your great work.

*/


/** Hardware specs for Atari's Dominos 

Memory Map:
		0000-03FF		RAM
		0400-07FF		DISPLAY RAM
		0800-0BFF	R	SWITCH
		0C00-0FFF	R	SYNC
		0C00-0C0F	W	ATTRACT
		0C10-0C1F	W	TUMBLE
		0C30-0C3F	W	LAMP2
		0C40-0C4F	W	LAMP1
		3000-37FF		Program ROM1
		3800-3FFF		Program ROM2
	   (F800-FFFF)		Program ROM2 - only needed for the 6502 vectors


*/


import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;
import java.net.*;
import java.applet.*;


class Dominos extends J6502Machine {

	/** Creator applet */

	J6502MachineApplet myCreator = null;


	/** Screen size */

	static int xScreen = 512;	// screen size
	static int yScreen = 448;


	// Timing for correct speed

	static	long timePerFrame  = 20; // Speed limit is 50 frames per second
	static	long thisFrameDuration = 0;
	static	long lastFrameTime = 0;


	/** Character set */

	int[] pixels;
	MemoryImageSource source;

	Image charSet; 


	/** Dirty buffer for Video RAM */
	int dirtyCounter = 0;
	int dirtyBuffer[] = new int[0x400*4];	// Buffer is 4 times the video ram to be
						// rather sure (CPU can 4 times change the

	/** VBLANK */

	private int VBLANK 		= 0xff;
	private boolean isVBLANK 	= false;


	// Colors: Alpha - Red - Green - Blue
	int BLACK	= 0xff000000;
	int LT_GREY	= 0xff808080;
	int DK_GREY	= 0xff555555;
	int WHITE	= 0xffffffff;
	


	/** Dominos creation */
	public Dominos ( J6502MachineApplet _parent ) {

		// *** Dominos runs at 750 KHz 
                super(_parent, 0.75, xScreen, yScreen );
		

		// *** Remember creator applet
		myCreator = _parent;
		
		// *** Set game directory
		GameDir = "dominos";


		// *** Define keys of input ports
		defineKey (KeyEvent.VK_UP,	PORT_ACTIVE_LOW, 0x0828, 0x80);
		defineKey (KeyEvent.VK_RIGHT,	PORT_ACTIVE_LOW, 0x0829, 0x80);
		defineKey (KeyEvent.VK_DOWN,	PORT_ACTIVE_LOW, 0x082a, 0x80);
		defineKey (KeyEvent.VK_LEFT,	PORT_ACTIVE_LOW, 0x082b, 0x80);
		defineKey (KeyEvent.VK_1,	PORT_ACTIVE_LOW, 0x082c, 0x80);  // Start 1
		defineKey (KeyEvent.VK_2,	PORT_ACTIVE_LOW, 0x082d, 0x80);  // Start 2
		defineKey (KeyEvent.VK_3,	PORT_ACTIVE_LOW, 0x0840, 0x40);  // Coin 1
		defineKey (KeyEvent.VK_4,	PORT_ACTIVE_LOW, 0x0840, 0x80);  // Coin 2

		// *** Initialize memory		
		for (int ii=0; ii<65536; ii++) mem[ii]=0xff;

		// *** Reset video RAM buffer
		dirtyCounter = 0;

		// *** Set DIP switches
		mem[0x811] = 0x80;	// 1 coin/1 credit
		mem[0x831] = 0x80;	// 1 coin/1 credit
		mem[0x815] = 0x80;	// 1 coin/1 credit
		mem[0x835] = 0x80;	// 1 coin/1 credit

	}

	public void handleKey( int keyAction, int keyCode) {
		if ( (keyAction == KeyEvent.KEY_PRESSED) && (keyCode == KeyEvent.VK_0 ) ) {
			myCreator.j6502machine.Debugging = true;
		}
	}



	/** Memory access */

	public void poke( int addr, int newByte ) {

		// *** Video RAM:
		if( (addr>=0x0400) & (addr<=0x07ff) ) {
			dirtyBuffer[dirtyCounter++] = addr-0x400; // remember dirty position			
			mem[addr]=newByte;
			return;
		}


                // *** Normal RAM:

		if ( addr < 0x3000 ) { 	// above this is Program ROM or nothing
			mem [addr]=newByte;	
		}
		
		return;
		
	}
	
	public int peek( int addr ) {

		// *** vertical sync with VBLANK

		if( addr==0x0c00 ) {
			if (isVBLANK) {
				isVBLANK = false;
				return(0x00);
			}
			else 	return(0xff);
		};


		// *** Normal memory:
		return mem[addr];
				
	}
	
	
	/****** Load cartridge file */
	public void loadRom () {

		// *** Load char prom into memory starting from 0x0000		
		loadRomChip (0x0000,0x0200,"7439-01.P4");	
		loadRomChip (0x0200,0x0200,"7440-01.R4");	

		// *** Load program proms into memory starting from 0x3000		
		loadRomChip (0x3000,0x0800,"7352-02.D1");	
		loadRomChip (0x3800,0x0800,"7438-02.E1");	
		loadRomChip (0xf800,0x0800,"7438-02.E1"); //Remapped for Reset, IRQ, and NMI vectors

		// *** Create Character set

		int pp, alt_color;

		pixels = new int[64*8*8*2*2 * 2];  // for 2 colors

		for (int j = 0; j < 4; j++) {
			for (int i = 0; i < 64*8; i++) {
				pp = (i%8)*4*64*8 + 2*8*((int)(i/8)) + j*2;
				if ( ((mem[i]>>(3-j))&0x01)==0 ) 	pixels[pp] = LT_GREY;
				else					pixels[pp] = BLACK;
				pixels[pp+1] = pixels[pp];

				// Second color
				if ( pixels[pp]==BLACK ) alt_color = WHITE;
				else			 alt_color = LT_GREY;

				pixels[pp+64*8*8*2*2]   = alt_color;
				pixels[pp+64*8*8*2*2+1] = alt_color;
			};
			for (int i = 0; i < 64*8; i++) {
				pp = (i%8)*4*64*8 + 2*8*((int)(i/8)) + (j+4)*2;
				if ( ((mem[i+0x200]>>(3-j))&0x01)==0 ) 	pixels[pp] = LT_GREY;
				else					pixels[pp] = BLACK;
				pixels[pp+1] = pixels[pp];

				// Second color
				if ( pixels[pp]==BLACK ) alt_color = WHITE;
				else			 alt_color = LT_GREY;

				pixels[pp+64*8*8*2*2]   = alt_color;
				pixels[pp+64*8*8*2*2+1] = alt_color;

			};
		};

		
		source = new MemoryImageSource(64*8*2, 8*2*2, pixels, 0, 64*8*2);
		charSet = parent.createImage(source);

	};



		

		
	/** Periodic handler to drive the 6502 interrupts*/

	public final int Loop6502() {

		loop6502Counter++;
		
		if (loop6502Counter % frameSkip == 0)  screenRefresh ();

		/* VBLANK port*/
		isVBLANK = true;

		return super.INT_NMI;

	}




	/** Screen drawing */	
	
	public void screenRefresh () {

		int i, m, pos, sx, sy, charcode, color;

		// *** Draw the character graphics

		for(i=0; i<dirtyCounter; i++) { 

			pos = dirtyBuffer[i];

			m = mem[0x0400+pos];
			color =	   (m & 0x80)>>3;
			charcode =  m & 0x3F;

                        sx = (pos % 32)<<4;
                        sy = ((int)(pos / 32))<<4;

			bufferGraphics.drawImage(charSet,sx,sy,sx+16,sy+16,
						(charcode<<4),color,
						(charcode<<4)+16, color+16,
						null);
		}

		dirtyCounter = 0;					// reset buffer

		canvasGraphics.drawImage(bufferImage,0,0,null);		// draw frame

		// Wait when too fast
		thisFrameDuration = System.currentTimeMillis() - lastFrameTime;
		if (thisFrameDuration < timePerFrame) {
			try { Thread.sleep(timePerFrame - thisFrameDuration); } 
			catch ( Exception ignored ) {}
		}
		lastFrameTime = System.currentTimeMillis();

	}


		
} // end of class Dominos


// ***
// *** Dominos Applet
// ***


public class JDominos extends J6502MachineApplet implements Runnable {
	
	// *** Create the Super breakout machine
	public J6502Machine createMachine ( ) {

		return ( new Dominos(this) ) ;

	}	
}



