/*********************************************************************************
 
	The Gravitar Emulator in Java

	by 

	Norbert Kehrer

	August 1998

**********************************************************************************

	Information sources:
		- Gravitar emulator of the MAME project - Thanks for the sources

	
	Gravitar is an Atari vector game using the AVG chip

**********************************************************************************


GRAVITAR SWITCH SETTINGS (Atari, 1982)
--------------------------------------

-------------------------------------------------------------------------------
Settings of 8-Toggle Switch on Gravitar PCB (at B4)
 8   7   6   5   4   3   2   1   Option
-------------------------------------------------------------------------------
Off On                           Free play
On  On                           1 coin for 2 credits
Off Off                          1 coin for 1 credit <
On  Off                          2 coins for 1 credit

        Off Off                  Right coin mechanism x 1 <
        On  Off                  Right coin mechanism x 4
        Off On                   Right coin mechanism x 5
        On  On                   Right coin mechanism x 6

                Off              Left coin mechanism x 1 <
                On               Left coin mechanism x 2

                    Off Off Off  No bonus coins <

                    Off On  Off  For every 4 coins inserted,
                                 logic adds 1 more coin
                    On  On  Off  For every 4 coins inserted,
                                 logic adds 2 more coins
                    Off Off On   For every 5 coins inserted,
                                 logic adds 1 more coin
                    On  Off On   For every 3 coins inserted,
                                 logic adds 1 more coin

                    Off On  On   No bonus coins
                    On  Off Off  ??? (not in manual!)
                    On  On  On   No bonus coins

-------------------------------------------------------------------------------

< Manufacturer's recommended setting

-------------------------------------------------------------------------------
Settings of 8-Toggle Switch on Gravitar PCB (at D4)
 8   7   6   5   4   3   2   1   Option
-------------------------------------------------------------------------------
                        On  On   No bonus
                        Off Off  Bonus ship every 10,000 points <
 d   d               d  On  Off  Bonus ship every 20,000 points
 e   e               e  Off On   Bonus ship every 30,000 points
 s   s               s
 U   U          On   U           Easy game play <
                Off              Hard game play
 t   t               t
 o   o  Off Off      o           3 ships per game
 N   N  On  Off      N           4 ships per game <
        Off On                   5 ships per game
        On  On                   6 ships per game

-------------------------------------------------------------------------------

< Manufacturer's recommended setting


*/


import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;
import java.net.*;
import java.applet.*;


// ***
// *** ATARI Gravitar Hardware
// ***

class Gravitar extends J6502Machine {

	// Video hardware
	
	public AtariAVG avg;		// Connected Atari vector processor

	static int xScreen = 512;	// screen size
	static int yScreen = 512;


	// Input ports 

	static int in_port_0 = 0x5f; // AVG not busy, No Service Mode
	static int in_port_3 = 0xff; // used for player controls

	
	// Timing for correct speed

	static	long timePerFrame  = 20; // Speed limit is 50 frames per second
	static	long thisFrameDuration = 0;
	static	long lastFrameTime = 0;


 
	/***** Gravitar creation ***/

	public Gravitar ( Container _parent ) {

		// *** Gravitar runs at 1.50 MHz  - we use 0.3 MHz for the 3 KHz source

                super( _parent, 0.3125, xScreen, yScreen );


		// *** Set game directory

		GameDir = "gravitar";


		// *** Clear EAROM (Hi scores)

		for (int i = 0; i < 0x40; i++) mem[0x8940+i] = 0x00;


		// *** Define keys of input ports

		defineKey (KeyEvent.VK_LEFT,	PORT_ACTIVE_LOW, 0x8000, 0x08); // Rotate Left
		defineKey (KeyEvent.VK_RIGHT,	PORT_ACTIVE_LOW, 0x8000, 0x04); // Rotate Right  
		defineKey (KeyEvent.VK_CONTROL,	PORT_ACTIVE_LOW, 0x8000, 0x02); // Fire 
		defineKey (KeyEvent.VK_UP,	PORT_ACTIVE_LOW, 0x8000, 0x10); // Thrust
		defineKey (KeyEvent.VK_SPACE,	PORT_ACTIVE_LOW, 0x8000, 0x01); // Hyperspace
		defineKey (KeyEvent.VK_3,	PORT_ACTIVE_LOW, 0x7800, 0x01); // Coin 1 
		defineKey (KeyEvent.VK_4,	PORT_ACTIVE_LOW, 0x7800, 0x02); // Coin 2 
		defineKey (KeyEvent.VK_1,	PORT_ACTIVE_LOW, 0x8000, 0x20); // Start 1 Player
		defineKey (KeyEvent.VK_2,	PORT_ACTIVE_LOW, 0x8000, 0x40); // Start 2 Players


		// *** Initialize memory	

		mem[0x7800] = in_port_0;


		// *** Pre-set DIP switch memory locations
	
		mem[0x8000] = 0xff;  
//		mem[0x8800] = 0xff;  



		// *** Create the AVG vector engine

		avg = new AtariAVG(mem, 0x2000 /* vector RAM */, bufferGraphics, xScreen, yScreen);

	}



	/***** Memory access ***/

	public final void poke( int addr, int newByte ) {


		switch ( addr ) {

		// *** Start the AVG engine and draw the screen

		case 0x8840:
			frameCounter++;
			if ((frameCounter % frameSkip) == 0) {
				avg.avgRun();
				canvasGraphics.drawImage(bufferImage,0,0,null);

			}

			// Wait when too fast
			thisFrameDuration = System.currentTimeMillis() - lastFrameTime;
			if (thisFrameDuration < timePerFrame) {
				try { Thread.sleep(timePerFrame - thisFrameDuration); } 
				catch ( Exception ignored ) {}
			}
			lastFrameTime = System.currentTimeMillis();
			break;

		case 0x8880:
			avg.avgReset();
			break;


		}


		// *** Normal RAM

		if ( addr < 0x9000 )         // 0x9000 - 0xffff ROM chips
			mem [addr]=newByte;	
		
		return;
		
	}
	

	public final int peek( int addr ) {

		// *** Pokey 1 Random Number Generator

		if ( addr == 0x600a ) return (int) (Math.random()*255);

		// *** Pokey 2 Random Number Generator

		if ( addr == 0x680a ) return (int) (Math.random()*255);


		if ( addr == 0x016c ) return 1;   // Fuel cheat
//			System.out.println("Read Fuel:"+mem[addr]);



		// *** Rest of memory

		return mem[addr];
	}
	
	
	/***** Load program ROMs ***/

	public void loadRom () {
		loadRomChip (0x2800,0x0800,"136010.210");  // vector ROM
		loadRomChip (0x3000,0x1000,"136010.207");  // vector ROM
		loadRomChip (0x4000,0x1000,"136010.208");  // vector ROM
		loadRomChip (0x5000,0x1000,"136010.209");  // vector ROM

		loadRomChip (0x9000,0x1000,"136010.201");  // code
		loadRomChip (0xa000,0x1000,"136010.202");  // code
		loadRomChip (0xb000,0x1000,"136010.203");  // code
		loadRomChip (0xc000,0x1000,"136010.204");  // code
		loadRomChip (0xd000,0x1000,"136010.205");  // code
		loadRomChip (0xe000,0x1000,"136010.206");  // code
		loadRomChip (0xf000,0x1000,"136010.206");  // reload code
	};


		
	/***** Periodic handler to drive the 6502 interrupts ***/

	public final int Loop6502() {
 	
		loop6502Counter++;
	
		// We update the 3 KHz Source

		in_port_0 = mem[0x7800];
		if ((in_port_0 & 0x80) == 0) 	in_port_0 |=  0x80;
		else 				in_port_0 &=  0x7f;
		mem[0x7800] = in_port_0;


		// We trigger an interrupt at a frequency of 240 Hz
		return super.INT_IRQ;
	}
	
		
} // end of class Gravitar



// ***
// *** ATARI Gravitar Applet
// ***


public class JGravitar extends J6502MachineApplet implements Runnable {

	// *** Create the Gravitar machine
	public J6502Machine createMachine ( ) {

		return ( new Gravitar(this) ) ;
	}

}



