/*********************************************************************************
 
	The Lunar Lander Emulator in Java

	by 

	Norbert Kehrer

	August 1998

**********************************************************************************

	Information sources:
		- Lunar Lander emulator of the MAME project - Thanks for the sources
	
	Lunar Lander is, I think,  the first Atari vector game from 1979

*/


import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;
import java.net.*;
import java.applet.*;


// ***
// *** ATARI Lunar Lander Hardware
// ***

class LunarLander extends J6502Machine {


	// Video hardware
	
	public AtariDVG dvg;		// Connected Atari vector processor

	static int xScreen = 512;	// screen size
	static int yScreen = 512;


	// Memory bank switch

	static int asteroid_bank = 0;
	static int asteroid_newbank;

	
	// Timing for correct speed

	static	long timePerFrame  = 20; // Speed limit is 50 frames per second
	static	long thisFrameDuration = 0;
	static	long lastFrameTime = 0;



	// Sounds

	static AudioClip beat1SFX       = null;
	static AudioClip beat2SFX       = null;
	static AudioClip largeExplSFX   = null;
	static AudioClip mediumExplSFX  = null;	
	static AudioClip smallExplSFX   = null;	
	static AudioClip largeSaucerSFX = null;
	static AudioClip smallSaucerSFX = null;
	static AudioClip thrustSFX      = null;
	static AudioClip ufoFireSFX     = null;
	static AudioClip yourFireSFX    = null;
	static AudioClip extraShipSFX   = null;

	static int lastPlayerFire, thisPlayerFire, lastUfoFire, thisUfoFire;	


 
	/***** Lunar Lander creation ***/

	public LunarLander ( Container _parent ) {

		// *** Lunar Lander runs at 1.50 Mhz  - but we use 0.3125 MHz for more exact NMIs

                super( _parent, 0.3125, xScreen, yScreen );
		

		// *** Set game directory

		GameDir = "llander";


		// *** Initialize memory	

		mem[0x2002] = 0; // Vector halt
		mem[0x2007] = 0; // Self test


		// *** Pre-set DIP switch memory locations
	
		mem[0x2800] = 0xfc | 0x02; // 2 = 1 Coin 1 Play,  0 = free play
		mem[0x2801] = 0xfc | 0x00; // Unknown
		mem[0x2802] = 0xfc | 0x01; // 4-ship game
		mem[0x2803] = 0xfc | 0x00; // English


		// *** Define keys of input ports

		defineKey (KeyEvent.VK_LEFT,	PORT_ACTIVE_HIGH, 0x2407, 0x80);
		defineKey (KeyEvent.VK_RIGHT,	PORT_ACTIVE_HIGH, 0x2406, 0x80);
		defineAnalogKey (KeyEvent.VK_UP,	0x2c00,  1, 10);
		defineAnalogKey (KeyEvent.VK_DOWN,	0x2c00, -1, 10);
		defineKey (KeyEvent.VK_CONTROL,	PORT_ACTIVE_HIGH, 0x2405, 0x80);
		defineKey (KeyEvent.VK_3,	PORT_ACTIVE_HIGH, 0x2401, 0x80);
		defineKey (KeyEvent.VK_4,	PORT_ACTIVE_HIGH, 0x2402, 0x80);
		defineKey (KeyEvent.VK_5,	PORT_ACTIVE_HIGH, 0x2403, 0x80);
		defineKey (KeyEvent.VK_1,	PORT_ACTIVE_HIGH, 0x2400, 0x80);
		defineKey (KeyEvent.VK_2,	PORT_ACTIVE_HIGH, 0x2404, 0x80);


		// *** Clear input port memory locations

		mem[0x2000] = 0xdf;
		for (int i = 0; i < 8; i++) {	
			mem[0x2400+i] = 0x7f;
		}


		// *** Create the DVG vector engine

		dvg = new AtariDVG(mem, 0x4000 /* vector RAM */, bufferGraphics, xScreen, yScreen);

	}



	/***** Memory access ***/

	public final void poke( int addr, int newByte ) {

		// *** Zero page is mirrored

		if ( (addr&0xff00) == 0 ) {
			mem [0x0100|addr]=newByte;
			return;
		}

		switch ( addr ) {

		// *** Start the DVG engine and draw the screen

		case 0x3000:
			frameCounter++;
			if ((frameCounter % frameSkip) == 0) {
				dvg.dvgRun();
				canvasGraphics.drawImage(bufferImage,0,0,null);
			}

			// Wait when too fast
			thisFrameDuration = System.currentTimeMillis() - lastFrameTime;
			if (thisFrameDuration < timePerFrame) {
				try { Thread.sleep(timePerFrame - thisFrameDuration); } 
				catch ( Exception ignored ) {}
			}
			lastFrameTime = System.currentTimeMillis();
			break;


		// *** Sound circuit control


		} // end case


		// *** Normal RAM

		if ( addr < 0x4800 )         // 0x4800 - 0xffff ROM chips
			mem [addr]=newByte;	
		
		return;
		
	}
	

	public final int peek( int addr ) {

		// *** Update the 3 KHz Source

		if ( addr == 0x2000 ) {
			if ((mem[0x2000]&0x40) != 0x00) 	mem[0x2000]&=0xdf;
			else 					mem[0x2000]|=0x40;
			return mem[addr];
		}


		// *** Zero page is mirrored

		if ( (addr&0xff00) == 0 ) 	return mem [0x0100|addr];
		else				return mem [addr];

	}



	/***** Load program ROMs ***/

	public void loadRom () {
		loadRomChip (0x6000,0x0800,"034572.02");  // code
		loadRomChip (0x6800,0x0800,"034571.02");  // code
		loadRomChip (0x7000,0x0800,"034570.02");  // code
		loadRomChip (0x7800,0x0800,"034569.02");  // code
		loadRomChip (0xf800,0x0800,"034569.02");  // reload code
		loadRomChip (0x4800,0x0800,"034599.01");  // vector ROM
		loadRomChip (0x5000,0x0800,"034598.01");  // vector ROM
		loadRomChip (0x5800,0x0800,"034597.01");  // vector ROM
	};


	/***** Load files for sound hardware ***/

	public void loadSound () {
/*
		beat1SFX =     	loadSoundFile ("Beat1.au");		
		beat2SFX =     	loadSoundFile ("Beat2.au");		
		largeExplSFX = 	loadSoundFile ("Largexpl.au");		
		mediumExplSFX =	loadSoundFile ("Medumxpl.au");		
		smallExplSFX =	loadSoundFile ("Smallxpl.au");		
		largeSaucerSFX= loadSoundFile ("Lsaucer.au");		
		smallSaucerSFX=	loadSoundFile ("Ssaucer.au");		
		thrustSFX = 	loadSoundFile ("Thrust.au");		
		ufoFireSFX = 	loadSoundFile ("Ufofire.au");		
		yourFireSFX = 	loadSoundFile ("Yourfire.au");		
		extraShipSFX = 	loadSoundFile ("Extraman.au");		
*/
	}



		
	/***** Periodic handler to drive the 6502 interrupts ***/

	public final int Loop6502() {

		loop6502Counter++;
	
		/*
		 * We catch the following busy loop:
		 * 6531 lda $2000
		 * 6534 lsr
		 * 6535 bcc 6531
		 */

		if ((PC >= 0x6531) && (PC <= 0x6536)) PC = 0x6537;


		// Interrupt

	 	return INT_NMI;
	}
		
} // end of class LunarLander



// ***
// *** ATARI Lunar Lander Applet
// ***


public class JLunarLander extends J6502MachineApplet implements Runnable {
	
	// *** Create the Lunar Lander machine
	public J6502Machine createMachine ( ) {

		return ( new LunarLander(this) ) ;

	}	
}



