/* 
        The Night Driver Emulator in Java

	by 

	Norbert Kehrer

	Written in June and July, 1999.

	Night Driver was an Atari arcade game from 1976.

        All the info from the MAME project (Mike Balfour's source code).
	Thank you for your great work.

*/


/** Hardware specs for Atari's Night Driver

Memory Map:
		0000-01FF	R/W 	SCRAM (Scratchpad RAM)
		0200-03FF	 W		PFW (Playfield Write)
		0400-05FF	 W		HVC (Horiz/Vert/Char for Roadway)
		0600-07FF	 R		IN0
		0800-09FF	 R		IN1
		0A00-0BFF	 W		OUT0
		0C00-0DFF	 W		OUT1
		0E00-0FFF	 -		OUT2 (Not used)
		8000-83FF	 R		PFR (Playfield Read)
		8400-87FF			Steering Reset
		8800-8FFF	 -		Spare (Not used)
		9000-97FF	 R		Program ROM1
		9800-9FFF	 R		Program ROM2
	   	(F800-FFFF)	 R		Program ROM2 - only needed for the 6502 vectors

*/


import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;
import java.net.*;
import java.applet.*;


class Nitedrvr extends J6502Machine {

	/** Creator applet */

	J6502MachineApplet myCreator = null;


	/** Screen size */

	static int xScreen = 512;	// screen size
	static int yScreen = 512;


	/** Character set */

	int[] pixels;
	MemoryImageSource source;

	Image charSet; 


	// Timing for correct speed

	static	long timePerFrame  = 20; // Speed limit is 50 frames per second
	static	long thisFrameDuration = 0;
	static	long lastFrameTime = 0;


	/** Dirty buffer for Video RAM */

	int dirtyCounter = 0;
	int dirtyBuffer[] = new int[0x400*4];	// Buffer is 4 times the video ram to be
						// rather sure (CPU can 4 times change the

	/** VBLANK */

	private int VBLANK 		= 0xff;
	private boolean isVBLANK 	= false;


	/** Gearing */

	int GEAR1 = 0xe0;
	int GEAR2 = 0xd0;
	int GEAR3 = 0xb0;
	int GEAR4 = 0x70;

	int gear   = GEAR1;	// actual gear (initially 1)
	int gearNr = 1;	


	/** Track selector */

	int track = 1;


	/** Colors: 	  Alpha - Red - Green - Blue */
	int BLACK	= 0xff000000;
	int WHITE	= 0xffffffff;
	


	/** Nitedrvr creation */

	public Nitedrvr ( J6502MachineApplet _parent ) {

		// *** Nitedrvr runs at 1 MHz 
                super(_parent, 1.00, xScreen, yScreen );
		

		// *** Remember creator applet
		myCreator = _parent;
		
		// *** Set game directory
		GameDir = "nitedrvr";


		// *** Define keys of input ports
		defineKey (KeyEvent.VK_RIGHT,	PORT_ACTIVE_HIGH, 0xe000, 0xc0);  // pseudo address	
		defineKey (KeyEvent.VK_LEFT,	PORT_ACTIVE_HIGH, 0xe000, 0x80);  // pseudo address	
		defineKey (KeyEvent.VK_1,	PORT_ACTIVE_LOW,  0xe001, 0x04);  // Start
		defineKey (KeyEvent.VK_UP,	PORT_ACTIVE_LOW,  0xe001, 0x08);  // Gas
		defineKey (KeyEvent.VK_3,	PORT_ACTIVE_LOW,  0xe001, 0x01);  // Coin 1
		defineKey (KeyEvent.VK_4,	PORT_ACTIVE_LOW,  0xe001, 0x02);  // Coin 2

		// *** Initialize memory		
		for (int ii=0; ii<65536; ii++) mem[ii]=0xff;

		// *** Reset video RAM buffer
		dirtyCounter = 0;

		// *** Set color "white" for roadway blocks
		canvasGraphics.setColor(Color.white);
		bufferGraphics.setColor(Color.white);
		buffer1Graphics.setColor(Color.white);

		// *** Set pseudo addresses for ports
		mem[0xe000] = 0x00;

	}

	public void handleKey( int keyAction, int keyCode) {
		if ( (keyAction == KeyEvent.KEY_PRESSED) && (keyCode == KeyEvent.VK_0 ) ) {
			myCreator.j6502machine.Debugging = true;
		}

		// *** Keys for the gearing

		if ( (keyAction == KeyEvent.KEY_PRESSED) && (keyCode == KeyEvent.VK_A ) ) {
			gear  = GEAR1;
			gearNr = 1;
		}
		if ( (keyAction == KeyEvent.KEY_PRESSED) && (keyCode == KeyEvent.VK_S ) ) {
			gear = GEAR2;
			gearNr = 2;
		}
		if ( (keyAction == KeyEvent.KEY_PRESSED) && (keyCode == KeyEvent.VK_D ) ) {
			gear = GEAR3;
			gearNr = 3;
		}
		if ( (keyAction == KeyEvent.KEY_PRESSED) && (keyCode == KeyEvent.VK_F ) ) {
			gear = GEAR4;
			gearNr = 4;
		}

		// *** Keys for the track selector

		if ( (keyAction == KeyEvent.KEY_PRESSED) && (keyCode == KeyEvent.VK_Q ) ) {
			track = 1;
		}
		if ( (keyAction == KeyEvent.KEY_PRESSED) && (keyCode == KeyEvent.VK_W ) ) {
			track = 2;
		}
		if ( (keyAction == KeyEvent.KEY_PRESSED) && (keyCode == KeyEvent.VK_E ) ) {
			track = 3;
		}

	}



	/** Memory access */

	public void poke( int addr, int newByte ) {

		// *** Video RAM:
		if( (addr>=0x0200) & (addr<=0x03ff) ) {
			dirtyBuffer[dirtyCounter++] = addr-0x200; // remember dirty position			
			mem[addr]=newByte;
			return;
		}


                // *** Normal RAM:

		if ( addr < 0x9000 ) { 	// above this is Program ROM or nothing
			mem [addr]=newByte;	
		}
		
		return;
		
	}
	
	public int peek( int addr ) {

		// *** Read input ports 0 and 1

		if ((addr>=0x0600) && (addr<=0x07ff)) return inPort0(addr);

		if ((addr>=0x0800) && (addr<=0x09ff)) return inPort1(addr);

		// *** Normal memory:
		return mem[addr];
				
	}
	
	
	/****** Load cartridge file */
	public void loadRom () {

		// *** Load char prom into memory starting from 0x0000		
		loadRomChip (0x0000,0x0200,"6568-01.P2");	

		// *** Load program proms into memory starting from 0x9000		
		loadRomChip (0x9000,0x0800,"6569-01.D2");	
		loadRomChip (0x9800,0x0800,"6570-01.F2");	
		loadRomChip (0xf800,0x0800,"6570-01.F2"); //Remapped for Reset, IRQ, and NMI vectors

		// *** Create Character set

		int pp, alt_color;

		pixels = new int[64*8*8*2*2]; 

		for (int c = 0; c < 64; c++) {
			for (int i = 0; i < 8; i++) {
				for (int j = 0; j < 8; j++) {
					pp = i*64*8*2*2 + c*8*2 + (7-j)*2;
					if ( ((mem[(c*8)+i]>>j)&0x01)==0 )
						pixels[pp] = BLACK;
					else	
						pixels[pp] = WHITE;
					pixels[pp+1] = pixels[pp];
				};
			};
		};

		source = new MemoryImageSource(64*8*2, 8*2, pixels, 0, 64*8*2);
		charSet = parent.createImage(source);

	};



		

		
	/** Periodic handler to drive the 6502 interrupts*/

	public final int Loop6502() {

		loop6502Counter++;
		
		if (loop6502Counter % frameSkip == 0)  screenRefresh ();

		/* VBLANK port*/
		isVBLANK = true;

		return super.INT_IRQ;

	}




	/** Screen drawing */	
	
	public void screenRefresh () {

		int i, j, pos, sx, sy, charcode, bx, by, ex, ey, x, y;

		// *** Draw the character graphics in buffer 1

		for(i=0; i<dirtyCounter; i++) { 

			pos = dirtyBuffer[i];

			charcode =  mem[0x0200+pos] & 0x3F;

                        sx = (pos % 32)<<4;
                        sy = ((int)(pos / 32))<<5;

			buffer1Graphics.drawImage(charSet,sx,sy,sx+16,sy+16,
					  	  (charcode<<4), 0,
						  (charcode<<4)+16, 16,
						  null);
		}

		dirtyCounter = 0;					// reset buffer

		bufferGraphics.drawImage(buffer1Image,0,0,null);	// draw buffer 1


		// *** Draw roadway in main buffer 
 
		for (i=0x0400; i < 0x0410; i++) {

			bx = mem[i]<<1;
			by = mem[i+16]<<1; 
			ex = bx + (((mem[i + 32] & 0xF0) >> 4)<<1);
			ey = by + ((16 - (mem[i + 32] & 0x0F))<<1);

			// Draw a block
			if ((bx!=ex) && (by!=ey))
				for (j=by; j<ey; j+=2) bufferGraphics.drawLine(bx, j, ex, j);
		}


		// *** Draw the "car backlight"

		bufferGraphics.fillRect(128,464,256,48);


		// *** Show info on gear and track - not a part of the original game

		bufferGraphics.drawString("Gear  "+gearNr, 10,464);
		bufferGraphics.drawString("Track "+track,  10,484);


		canvasGraphics.drawImage(bufferImage,0,0,null);		// draw frame

		// Wait when too fast
		thisFrameDuration = System.currentTimeMillis() - lastFrameTime;
		if (thisFrameDuration < timePerFrame) {
			try { Thread.sleep(timePerFrame - thisFrameDuration); } 
			catch ( Exception ignored ) {}
		}
		lastFrameTime = System.currentTimeMillis();

	}



	// *** Now the port methods - all info from Mike Balfour's MAME driver source code

	/***************************************************************************
	Steering
	
	When D7 is high, the steering wheel has moved.
	If D6 is low, it moved left.  If D6 is high, it moved right.
	Be sure to keep returning a direction until steering_reset is called,
	because D6 and D7 are apparently checked at different times, and a
	change in-between can affect the direction you move.
	***************************************************************************/

	/***************************************************************************
	Input port 0

	Night Driver looks for the following:
		A: $00
			D4 - OPT1
			D5 - OPT2
			D6 - OPT3
			D7 - OPT4
		A: $01
			D4 - TRACK SET
			D5 - BONUS TIME ALLOWED
			D6 - VBLANK
			D7 - !TEST
		A: $02
			D4 - !GEAR 1
			D5 - !GEAR 2
			D6 - !GEAR 3
			D7 - SPARE
		A: $03
			D4 - SPARE
			D5 - DIFFICULT BONUS
			D6 - STEER A
			D7 - STEER B

	Fill in the steering and gear bits in a special way.
	***************************************************************************/


	final int inPort0 (int addr) {

		switch (addr & 0x03) {
			case 0x00:
				return 0x41;		// DIP switch: 1 play/coin, 75 seconds
			case 0x01:
				return 0xa0;		// DIP switch: Normal track, bonus time
			case 0x02:
				return gear;		// actual gear
			case 0x03:
				return mem[0xe000];	// pseudo address for steering
			default:
				return 0xff;
		}
	}

	/***************************************************************************

	Night Driver looks for the following:
		A: $00
			D6 - SPARE
			D7 - COIN 1
		A: $01
			D6 - SPARE
			D7 - COIN 2
		A: $02
			D6 - SPARE
			D7 - !START
		A: $03
			D6 - SPARE
			D7 - !ACC
		A: $04
			D6 - SPARE
			D7 - EXPERT
		A: $05
			D6 - SPARE
			D7 - NOVICE
		A: $06
			D6 - SPARE
			D7 - Special Alternating Signal
		A: $07
			D6 - SPARE
			D7 - Ground

	Fill in the track difficulty switch and special signal in a special way.
	***************************************************************************/


	final int inPort1 (int addr) {

		int port;

		port = mem[0xe001];

		switch (addr & 0x07) {
			case 0x00:
				return ((port & 0x01) << 7);
			case 0x01:
				return ((port & 0x02) << 6);
			case 0x02:
				return ((port & 0x04) << 5);
			case 0x03:
				return ((port & 0x08) << 4);
			case 0x04:
				if (track == 2) return 0x80; else return 0x00;
			case 0x05:
				if (track == 1) return 0x80; else return 0x00;
			case 0x06:
				// TODO: fix alternating signal? 
			case 0x07:
				return 0x00;
			default:
				return 0xFF;
		}
	}


		
} // end of class Nitedrvr


// ***
// *** Night Driver Applet
// ***


public class JNitedrvr extends J6502MachineApplet implements Runnable {
	
	// *** Create the Super breakout machine
	public J6502Machine createMachine ( ) {

		return ( new Nitedrvr(this) ) ;

	}	
}



