/*********************************************************************************
 
	The New Atari DVG Vector Processor Emulator in Java

	by 

	Norbert Kehrer

	December 1999

**********************************************************************************

	Information sources:
		- Asteroids emulator of the MAME project - Thanks for the sources
		- Asteroids emulator by Chris Pile       - Thanks for the documentation

	
	The Atari DVG is a vector engine used in Atari games like Asteroids

	This engine supports new graphics for Asteroids.

*/


import java.awt.*;
import java.awt.image.*;
import java.awt.image.MemoryImageSource;
import java.awt.image.ColorModel;
import java.util.*;
import java.io.*;
import java.net.*;
import java.applet.*;
import java.net.*;


// ***
// *** ATARI DVG vector engine
// ***

public class NAtariDVG extends Object {

	/* Data for graphically enhanced objects */

	static NAsteroids myAsteroids;		// Machine
	static boolean vectorDrawing = true;	// Switch to stop vectors for g.e.objects


	/* DVG memory and video hardware information */
	public int[] 	dvgMem;
	public int	dvgVectorRAM;
	public Graphics	dvgBufferGraphics;

	static int xSize;
	static int ySize;
	
	static Color[] dvgColorTable = new Color[16];


	/* DVG stack size */
	static int MAXSTACK = 8;

	/* DVG commands */
	static int DVCTR = 0x01;
	static int DLABS = 0x0a;
	static int DHALT = 0x0b;
	static int DJSRL = 0x0c;
	static int DRTSL = 0x0d;
	static int DJMPL = 0x0e;
	static int DSVEC = 0x0f;


	/* Chip Registers */
	static int pc;
	static int sp;
	static int[] stack = new int [MAXSTACK];

	static int scale;
	static int currentx, currenty;



	// *** DVG generation 

	public NAtariDVG (int [] memory, int vectorRAM, Graphics bufferGr, int xs, int ys, NAsteroids ast) {
		dvgMem = 		memory;
		dvgVectorRAM = 		vectorRAM;
		dvgBufferGraphics = 	bufferGr;
		xSize =			xs;
		ySize =			ys;
		myAsteroids =		ast;

		// Generate colors
		for (int i = 0; i < 16; i++) dvgColorTable[i] = new Color(i<<4, i<<4, i<<4);

		// Set drawing mode
		vectorDrawing = true;
	}
	
	
	// *** Utility functions
	
	final int readWordVectorRAM (int offset ) {
		int dvgAddr = dvgVectorRAM + (offset<<1);
		return ( dvgMem[dvgAddr] | ( dvgMem[dvgAddr+1] << 8) );
	}

	final int twosCompVal( int num, int bits) {
		num &= 0x0fff;
		return ( ((num&0x0800) != 0) 	? (num | ~(0x0fff))
						: (num & 0xfff)
		);
	}



	// *** Vector drawing

	final void drawVector(int x, int y, int dx, int dy, int intensity) {
		int x1,y1,x2,y2,corr=3;
		if ((intensity > 0) && vectorDrawing) {
			x1 = (x>>corr);
			y1 = xSize-(y>>corr);
			x2 = ((x+dx)>>corr);
			y2 = ySize-((y+dy)>>corr);
			dvgBufferGraphics.setColor(dvgColorTable[intensity>>4]);
			dvgBufferGraphics.drawLine(x1, y1, x2, y2);
		}
	}


	// *** Draw graphically enhanced objects

	final void drawObj(int addr, int x, int y, int scale) {
		int	x1,y1,dx,dy, x2,y2,corr=3;
		Image	newImg;

		x1 = (x>>corr);
		y1 = xSize-(y>>corr);

		dx = dy = 0;
		newImg = null;

		// sc 0 = l, 15 = m, 14 = s

		switch(addr) {

		case 0x08f4:	// *** rock 1
			switch (scale) {
			case 0:
				newImg =	myAsteroids.rock1L;
				dx = 		16;
				dy =		16;
				break;
			case 15:
				newImg =	myAsteroids.rock1M;
				dx = 		8;
				dy =		8;
				break;
			case 14:
				newImg =	myAsteroids.rock1S;
				dx = 		4;
				dy =		4;
				break;
			}
			break;

		case 0x0900:	// *** rock 2
			switch (scale) {
			case 0:
				newImg =	myAsteroids.rock1L;
				dx = 		16;
				dy =		16;
				break;
			case 15:
				newImg =	myAsteroids.rock1M;
				dx = 		8;
				dy =		8;
				break;
			case 14:
				newImg =	myAsteroids.rock1S;
				dx = 		4;
				dy =		4;
				break;
			}
			break;

		case 0x090e:	// *** rock 3
			switch (scale) {
			case 0:
				newImg =	myAsteroids.rock1L;
				dx = 		16;
				dy =		16;
				break;
			case 15:
				newImg =	myAsteroids.rock1M;
				dx = 		8;
				dy =		8;
				break;
			case 14:
				newImg =	myAsteroids.rock1S;
				dx = 		4;
				dy =		4;
				break;
			}
			break;

		case 0x091b:	// *** rock 4
			switch (scale) {
			case 0:
				newImg =	myAsteroids.rock1L;
				dx = 		16;
				dy =		16;
				break;
			case 15:
				newImg =	myAsteroids.rock1M;
				dx = 		8;
				dy =		8;
				break;
			case 14:
				newImg =	myAsteroids.rock1S;
				dx = 		4;
				dy =		4;
				break;
			}
			break;

		case 0x092a:	// *** saucer
			switch (scale) {
			case 15:
				newImg =	myAsteroids.saucerL;
				dx = 		10;
				dy =		 6;
				break;
			case 14:
				newImg =	myAsteroids.saucerS;
				dx = 		5;
				dy =		3;
				break;
			}
			break;

		case 0x0ade:	// *** letter O
			if (scale != 0) {
				newImg =	myAsteroids.letterO;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0b2f:	// *** letter 1
			if (scale != 0) {
				newImg =	myAsteroids.letter1;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0b33:	// *** letter 2
			if (scale != 0) {
				newImg =	myAsteroids.letter2;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0b3b:	// *** letter 3
			if (scale != 0) {
				newImg =	myAsteroids.letter3;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0b42:	// *** letter 4
			if (scale != 0) {
				newImg =	myAsteroids.letter4;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0b49:	// *** letter 5
			if (scale != 0) {
				newImg =	myAsteroids.letter5;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0b50:	// *** letter 6
			if (scale != 0) {
				newImg =	myAsteroids.letter6;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0b57:	// *** letter 7
			if (scale != 0) {
				newImg =	myAsteroids.letter7;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0b5c:	// *** letter 8
			if (scale != 0) {
				newImg =	myAsteroids.letter8;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0b64:	// *** letter 9
			if (scale != 0) {
				newImg =	myAsteroids.letter9;
				dx = 		0;
				dy =		13;
			}
			break;


		case 0x0a79:	// *** letter A
			if (scale != 0) {
				newImg =	myAsteroids.letterA;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0a81:	// *** letter B
			if (scale != 0) {
				newImg =	myAsteroids.letterB;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0a8e:	// *** letter C
			if (scale != 0) {
				newImg =	myAsteroids.letterC;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0a94:	// *** letter D
			if (scale != 0) {
				newImg =	myAsteroids.letterD;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0a9c:	// *** letter E
			if (scale != 0) {
				newImg =	myAsteroids.letterE;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0aa4:	// *** letter F
			if (scale != 0) {
				newImg =	myAsteroids.letterF;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0aab:	// *** letter G
			if (scale != 0) {
				newImg =	myAsteroids.letterG;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0ab4:	// *** letter H
			if (scale != 0) {
				newImg =	myAsteroids.letterH;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0abb:	// *** letter I
			if (scale != 0) {
				newImg =	myAsteroids.letterI;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0ac2:	// *** letter J
			if (scale != 0) {
				newImg =	myAsteroids.letterJ;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0ac8:	// *** letter K
			if (scale != 0) {
				newImg =	myAsteroids.letterK;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0ace:	// *** letter L
			if (scale != 0) {
				newImg =	myAsteroids.letterL;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0ad3:	// *** letter M
			if (scale != 0) {
				newImg =	myAsteroids.letterM;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0ad9:	// *** letter N
			if (scale != 0) {
				newImg =	myAsteroids.letterN;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0ae4:	// *** letter P
			if (scale != 0) {
				newImg =	myAsteroids.letterP;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0aeb:	// *** letter Q
			if (scale != 0) {
				newImg =	myAsteroids.letterQ;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0af4:	// *** letter R
			if (scale != 0) {
				newImg =	myAsteroids.letterR;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0afc:	// *** letter S
			if (scale != 0) {
				newImg =	myAsteroids.letterS;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0b03:	// *** letter T
			if (scale != 0) {
				newImg =	myAsteroids.letterT;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0b09:	// *** letter U
			if (scale != 0) {
				newImg =	myAsteroids.letterU;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0b0f:	// *** letter V
			if (scale != 0) {
				newImg =	myAsteroids.letterV;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0b14:	// *** letter W
			if (scale != 0) {
				newImg =	myAsteroids.letterW;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0b1b:	// *** letter X
			if (scale != 0) {
				newImg =	myAsteroids.letterX;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0b20:	// *** letter Y
			if (scale != 0) {
				newImg =	myAsteroids.letterY;
				dx = 		0;
				dy =		13;
			}
			break;

		case 0x0b27:	// *** letter Z
			if (scale != 0) {
				newImg =	myAsteroids.letterZ;
				dx = 		0;
				dy =		13;
			}
			break;



		case 0x08d1:	// *** explosion1
				newImg =	myAsteroids.expl1;
				dx = 		13;
				dy =		13;
			break;

		case 0x08b6:	// *** explosion2
				newImg =	myAsteroids.expl2;
				dx = 		27;
				dy =		27;
			break;

		case 0x0897:	// *** explosion3
				newImg =	myAsteroids.expl3;
				dx = 		38;
				dy =		34;
			break;

		case 0x0881:	// *** explosion4
				newImg =	myAsteroids.expl4;
				dx = 		39;
				dy =		36;
			break;

		}



		// Draw new object if found and turn off vectors
		if ( newImg != null ) {
			vectorDrawing = false;
			dvgBufferGraphics.drawImage(newImg,x1-dx,y1-dy,null);
		}

	}



	// *** Execute the DVG commands
	
	public final void dvgRun() {

		boolean done = false;

		int firstwd  = 0;
		int secondwd = 0;
		int opcode;

		int x, y;
		int z, temp;
		int a;

		int deltax, deltay;


		// *** Clear Screen, i.e. draw background

		// dvgBufferGraphics.setColor (Color.black);
		// dvgBufferGraphics.fillRect(0,0,xSize,ySize);

		dvgBufferGraphics.drawImage(myAsteroids.background1,0,0,null);


		// *** Initialize state machine

		pc = 0;
		sp = 0;
		scale = 0;
		
		currentx = 0;
		currenty = 0;


		// *** Execute the DVG commands

		while (!done) {

			firstwd = readWordVectorRAM(pc);
			opcode = firstwd >> 12;
			
			pc++;

			if ((opcode >= DVCTR ) && (opcode <= DLABS)) {
				secondwd = readWordVectorRAM (pc);
				pc++;
			}


			switch (opcode) {

			case 0: // DHALT
				done = true;
				break;
			case 1:
			case 2:
			case 3:
			case 4:
			case 5:
			case 6:
			case 7:
			case 8:
			case 9: // DLVEC
	  			y = firstwd & 0x03ff;
				if ((firstwd & 0x0400) != 0) y = -y;
				x = secondwd & 0x03ff;
				if ((secondwd & 0x0400) != 0)	x = -x;
				z = secondwd >> 12;

	  			temp = ((scale + opcode) & 0x0f);

	  			if (temp > 9) temp = -1;
	  			deltax = x >> (7-temp);
				deltay = y >> (7-temp);

				if (z != 0) z = (z << 4) | 0x0f;

				drawVector(currentx, currenty, deltax, deltay, z); 

	  			currentx += deltax;
				currenty += deltay;
								
				break;

			case 0x0a: // DLABS
				x = twosCompVal (secondwd, 12);
				y = twosCompVal (firstwd, 12);
	  			scale = (secondwd >> 12);
				currentx = x<<2;
				currenty = y<<2;

				break;

			case 0x0b: // DHALT
				done = true;
				break;

			case 0x0c: // DJSRL
				a = firstwd & 0x0fff;
			
				stack [sp] = pc;
				if (sp == (MAXSTACK - 1)) {
					System.out.println("\n*** Vector generator stack overflow! ***\n");
					done = true;
					sp = 0;
				}
				else
					sp++;
					
				pc = a;

				// Draw new graphics
				drawObj(a, currentx, currenty, scale);

				break;

			case 0x0d: // DRTSL
				if (sp == 0) {
					System.out.println("\n*** Vector generator stack overflow! ***\n");
					done = true;
					sp = MAXSTACK - 1;
				}
				else
					sp--;
					
				pc = stack [sp];

				// Turn on vectors again after drawing new graphics
				if (!vectorDrawing) vectorDrawing = true;

				break;

			case 0x0e: // DJMPL
				a = firstwd & 0x0fff;
				pc = a;
				break;

			case 0x0f: // DSVEC
				y = firstwd & 0x0300;
				if ((firstwd & 0x0400) != 0) y = -y;
				x = (firstwd & 0x03) << 8;
				if ((firstwd & 0x04) != 0) x = -x;
				z = (firstwd >> 4) & 0x0f;
				temp = 2 + ((firstwd >> 2) & 0x02) + ((firstwd >>11) & 0x01);
	  			temp = ((scale + temp) & 0x0f);

				if (temp > 9) temp = -1;

				deltax = x >> (7-temp);
				deltay = y >> (7-temp);
	  			
				if (z != 0) z = (z << 4) | 0x0f;

				drawVector(currentx, currenty, deltax, deltay, z); 

				currentx += deltax;
				currenty += deltay;
				
				break;

			default:
				System.out.println("Unknown DVG opcode found!");
				done = true;
			}
		}


	}
}








