/*********************************************************************************
 
          The New Asteroids Emulator in Java

	by 

	Norbert Kehrer

	August 1998

**********************************************************************************

	Information sources:
		- Asteroids emulator of the MAME project - Thanks for the sources
		- Asteroids emulator by Chris Pile       - Thanks for the documentation

	
	ASTEROIDS is a famous Atari vector game from 1979

*/


import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;
import java.net.*;
import java.applet.*;


// ***
// *** ATARI Asteroids Hardware
// ***

class Asteroids extends J6502Machine {


	// Video hardware
	
	public AtariDVG dvg;		// Connected Atari vector processor

	static int xScreen = 512;	// screen size
	static int yScreen = 512;


	// Memory bank switch

	static int asteroid_bank = 0;
	static int asteroid_newbank;

	
	// Timing for correct speed

	static	long timePerFrame  = 20; // Speed limit is 50 frames per second
	static	long thisFrameDuration = 0;
	static	long lastFrameTime = 0;



	// Sounds

	static AudioClip beat1SFX       = null;
	static AudioClip beat2SFX       = null;
	static AudioClip largeExplSFX   = null;
	static AudioClip mediumExplSFX  = null;	
	static AudioClip smallExplSFX   = null;	
	static AudioClip largeSaucerSFX = null;
	static AudioClip smallSaucerSFX = null;
	static AudioClip thrustSFX      = null;
	static AudioClip ufoFireSFX     = null;
	static AudioClip yourFireSFX    = null;
	static AudioClip extraShipSFX   = null;

	static int lastPlayerFire, thisPlayerFire, lastUfoFire, thisUfoFire;	


 
	/***** Asteroids creation ***/

	public Asteroids ( Container _parent ) {

		// *** Asteroids runs at 1.50 Mhz  - but we use 0.3125 MHz for more exact NMIs

                super( _parent, 0.3125, xScreen, yScreen );
		

		// *** Set game directory

		GameDir = "asteroid";


		// *** Initialize memory	

		mem[0x2002] = 0; // Vector halt
		mem[0x2007] = 0; // Self test


		// *** Pre-set DIP switch memory locations
	
		mem[0x2800] = 0xfc | 0x02; // 2 = 1 Coin 1 Play,  0 = free play
		mem[0x2801] = 0xfc | 0x00; // Unknown
		mem[0x2802] = 0xfc | 0x01; // 4-ship game
		mem[0x2803] = 0xfc | 0x00; // English


		// *** Define keys of input ports

		defineKey (KeyEvent.VK_LEFT,	PORT_ACTIVE_HIGH, 0x2407, 0x80);
		defineKey (KeyEvent.VK_RIGHT,	PORT_ACTIVE_HIGH, 0x2406, 0x80);
		defineKey (KeyEvent.VK_CONTROL,	PORT_ACTIVE_HIGH, 0x2004, 0x80);
		defineKey (KeyEvent.VK_UP,	PORT_ACTIVE_HIGH, 0x2405, 0x80);
		defineKey (KeyEvent.VK_SPACE,	PORT_ACTIVE_HIGH, 0x2003, 0x80);
		defineKey (KeyEvent.VK_3,	PORT_ACTIVE_HIGH, 0x2400, 0x80);
		defineKey (KeyEvent.VK_4,	PORT_ACTIVE_HIGH, 0x2401, 0x80);
		defineKey (KeyEvent.VK_5,	PORT_ACTIVE_HIGH, 0x2402, 0x80);
		defineKey (KeyEvent.VK_1,	PORT_ACTIVE_HIGH, 0x2403, 0x80);
		defineKey (KeyEvent.VK_2,	PORT_ACTIVE_HIGH, 0x2404, 0x80);


		// *** Clear input port memory locations

		for (int i = 0; i < 8; i++) {	
			mem[0x2000+i] = 0x7f;
			mem[0x2400+i] = 0x7f;
		}


		// *** Create the DVG vector engine

		dvg = new AtariDVG(mem, 0x4000 /* vector RAM */, bufferGraphics, xScreen, yScreen);

	}



	/***** Memory access ***/

	public final void poke( int addr, int newByte ) {

		switch ( addr ) {

		// *** Start the DVG engine and draw the screen

		case 0x3000:
			frameCounter++;
			if ((frameCounter % frameSkip) == 0) {
				dvg.dvgRun();
				canvasGraphics.drawImage(bufferImage,0,0,null);
			}

			// Wait when too fast
			thisFrameDuration = System.currentTimeMillis() - lastFrameTime;
			if (thisFrameDuration < timePerFrame) {
				try { Thread.sleep(timePerFrame - thisFrameDuration); } 
				catch ( Exception ignored ) {}
			}
			lastFrameTime = System.currentTimeMillis();
			break;


		// *** Memory bank switch

		case 0x3200:
			asteroid_newbank = (newByte >> 2) & 0x01;
			if (asteroid_bank != asteroid_newbank) {
				/* Perform bankswitching on page 2 and page 3 */
				int temp;
				int i;
				asteroid_bank = asteroid_newbank;
				for (i = 0; i < 0x100; i++) {
					temp = mem[0x200 + i];
					mem[0x200 + i] = mem[0x300 + i];
					mem[0x300 + i] = temp;
				}
			}
			break;


		// *** Sound circuit control

		case 0x3600: // Explosions
			if (newByte == 0x3d) largeExplSFX.play(); 
			if (newByte == 0xfd) mediumExplSFX.play(); 
			if (newByte == 0xbd) smallExplSFX.play(); 
			break;

		case 0x3a00: // Heartbeat
			if ((newByte&0x10) != 0) {
				if ((newByte&0x04) != 0) beat2SFX.play(); 
				else			 beat1SFX.play(); 
			}
			break;	

		case 0x3c00: // Saucer
			if ((newByte&0x80) != 0) {
				if ((mem[0x3c02]&0x80) != 0)	 largeSaucerSFX.loop();	
				else				 smallSaucerSFX.loop();	
			}	
			else {
				largeSaucerSFX.stop();	
				smallSaucerSFX.stop();	
			}
 			break;	
		
		case 0x3c01: // Ufo fire
			thisUfoFire = (newByte&0x80);
			if (thisUfoFire != 0) 
				if (thisUfoFire != lastUfoFire) ufoFireSFX.play();	
			lastUfoFire = thisUfoFire;	
 			break;
		
		case 0x3c03: // Thrust
			if ((newByte&0x80) != 0) thrustSFX.loop();	
			else			 thrustSFX.stop();	
 			break;	

		case 0x3c04: // Player fire
			thisPlayerFire = (newByte&0x80);
			if (thisPlayerFire != 0) 
				if (thisPlayerFire != lastPlayerFire) yourFireSFX.play();	
			lastPlayerFire = thisPlayerFire;	
 			break;	

		case 0x3c05: // Extra ship
			if (newByte != 0) 	extraShipSFX.loop();	
			else			extraShipSFX.stop();	
 			break;

		} // end case


		// *** Normal RAM

		if ( addr < 0x6800 )         // 0x6800 - 0xffff ROM chips
			mem [addr]=newByte;	
		
		return;
		
	}
	

	
	/***** Load program ROMs ***/

	public void loadRom () {
		loadRomChip (0x6800,0x0800,"035145.02");  // code
		loadRomChip (0x7000,0x0800,"035144.02");  // code
		loadRomChip (0x7800,0x0800,"035143.02");  // code
		loadRomChip (0xf800,0x0800,"035143.02");  // reload code
		loadRomChip (0x5000,0x0800,"035127.02");  // vector ROM
	};


	/***** Load files for sound hardware ***/

	public void loadSound () {

		beat1SFX =     	loadSoundFile ("Beat1.au");		
		beat2SFX =     	loadSoundFile ("Beat2.au");		
		largeExplSFX = 	loadSoundFile ("Largexpl.au");		
		mediumExplSFX =	loadSoundFile ("Medumxpl.au");		
		smallExplSFX =	loadSoundFile ("Smallxpl.au");		
		largeSaucerSFX= loadSoundFile ("Lsaucer.au");		
		smallSaucerSFX=	loadSoundFile ("Ssaucer.au");		
		thrustSFX = 	loadSoundFile ("Thrust.au");		
		ufoFireSFX = 	loadSoundFile ("Ufofire.au");		
		yourFireSFX = 	loadSoundFile ("Yourfire.au");		
		extraShipSFX = 	loadSoundFile ("Extraman.au");		
	}



		
	/***** Periodic handler to drive the 6502 interrupts ***/

	public final int Loop6502() {

		loop6502Counter++;
	
		/*
		 * We catch the following busy loop:
		 * 6812 lda $2002
		 * 6815 bmi $6812
		 */

		if ((PC >= 0x6812) && (PC <= 0x6816)) PC = 0x6817;


		// Interrupt

	 	return INT_NMI;
	}
		
} // end of class Asteroids



// ***
// *** ATARI Asteroids Applet
// ***


public class JAsteroids extends J6502MachineApplet implements Runnable {
	
	// *** Create the Asteroids machine
	public J6502Machine createMachine ( ) {

		return ( new Asteroids(this) ) ;

	}	
}



